"""
Helpers to support application of Jacobians to vectors
Helpers to check input and output sizes of Jacobian-matrix products.
"""
import functools


###############################################################################
#                              Utility functions                              #
###############################################################################
def add_V_dim(mat):
    return mat.unsqueeze(0)


def remove_V_dim(mat):
    if mat.shape[0] != 1:
        raise RuntimeError(
            "Cannot unsqueeze dimension 0. ", "Got tensor of shape {}".format(mat.shape)
        )
    return mat.squeeze(0)


def check_shape(mat, like, diff=1):
    """Compare dimension diff,diff+1, ... with dimension 0,1,..."""
    mat_shape = [int(dim) for dim in mat.shape]
    like_shape = [int(dim) for dim in like.shape]

    if len(mat_shape) - len(like_shape) != diff:
        raise RuntimeError(
            "Difference in dimension must be {}.".format(diff),
            " Got {} and {}".format(mat_shape, like_shape),
        )
    if mat_shape[diff:] != like_shape:
        raise RuntimeError(
            "Compared shapes {} and {} do not match. ".format(
                mat_shape[diff:], like_shape
            ),
            "Got {} and {}".format(mat_shape, like_shape),
        )


def check_same_V_dim(mat1, mat2):
    V1, V2 = mat1.shape[0], mat2.shape[0]
    if V1 != V2:
        raise RuntimeError("Number of vectors changed. Got {} and {}".format(V1, V2))


def check_like(mat, module, name, diff=1, *args, **kwargs):
    return check_shape(mat, getattr(module, name), diff=diff)


def check_like_with_sum_batch(mat, module, name, sum_batch=True, *args, **kwargs):
    diff = 1 if sum_batch else 2
    return check_shape(mat, getattr(module, name), diff=diff)


def same_dim_as(mat, module, name, *args, **kwargs):
    return len(mat.shape) == len(getattr(module, name).shape)


###############################################################################
#            Decorators for handling vectors as matrix special case           #
###############################################################################
def mat_prod_accept_vectors(mat_prod, vec_criterion):
    """Add support for vectors to matrix products.

    vec_criterion(mat, module) returns if mat is a vector.
    """

    @functools.wraps(mat_prod)
    def wrapped_mat_prod_accept_vectors(
        self, module, g_inp, g_out, mat, *args, **kwargs
    ):
        is_vec = vec_criterion(mat, module, *args, **kwargs)
        mat_in = mat if not is_vec else add_V_dim(mat)
        mat_out = mat_prod(self, module, g_inp, g_out, mat_in, *args, **kwargs)
        mat_out = mat_out if not is_vec else remove_V_dim(mat_out)

        return mat_out

    return wrapped_mat_prod_accept_vectors


# vec criteria
same_dim_as_output = functools.partial(same_dim_as, name="output")
same_dim_as_input = functools.partial(same_dim_as, name="input0")
same_dim_as_weight = functools.partial(same_dim_as, name="weight")
same_dim_as_bias = functools.partial(same_dim_as, name="bias")

# decorators for handling vectors
jac_t_mat_prod_accept_vectors = functools.partial(
    mat_prod_accept_vectors, vec_criterion=same_dim_as_output,
)

weight_jac_t_mat_prod_accept_vectors = functools.partial(
    mat_prod_accept_vectors, vec_criterion=same_dim_as_output,
)
bias_jac_t_mat_prod_accept_vectors = functools.partial(
    mat_prod_accept_vectors, vec_criterion=same_dim_as_output,
)

jac_mat_prod_accept_vectors = functools.partial(
    mat_prod_accept_vectors, vec_criterion=same_dim_as_input,
)

weight_jac_mat_prod_accept_vectors = functools.partial(
    mat_prod_accept_vectors, vec_criterion=same_dim_as_weight,
)

bias_jac_mat_prod_accept_vectors = functools.partial(
    mat_prod_accept_vectors, vec_criterion=same_dim_as_bias,
)


###############################################################################
#       Decorators for checking inputs and outputs of mat_prod routines       #
###############################################################################
def mat_prod_check_shapes(mat_prod, in_check, out_check):
    """Check that input and output have correct shapes."""

    @functools.wraps(mat_prod)
    def wrapped_mat_prod_check_shapes(self, module, g_inp, g_out, mat, *args, **kwargs):
        in_check(mat, module, *args, **kwargs)
        mat_out = mat_prod(self, module, g_inp, g_out, mat, *args, **kwargs)
        out_check(mat_out, module, *args, **kwargs)
        check_same_V_dim(mat_out, mat)

        return mat_out

    return wrapped_mat_prod_check_shapes


# input/output checker
shape_like_output = functools.partial(check_like, name="output")
shape_like_input = functools.partial(check_like, name="input0")
shape_like_weight = functools.partial(check_like, name="weight")
shape_like_bias = functools.partial(check_like, name="bias")
shape_like_weight_with_sum_batch = functools.partial(
    check_like_with_sum_batch, name="weight"
)
shape_like_bias_with_sum_batch = functools.partial(
    check_like_with_sum_batch, name="bias"
)

# decorators for shape checking
jac_mat_prod_check_shapes = functools.partial(
    mat_prod_check_shapes, in_check=shape_like_input, out_check=shape_like_output
)

weight_jac_mat_prod_check_shapes = functools.partial(
    mat_prod_check_shapes, in_check=shape_like_weight, out_check=shape_like_output
)

bias_jac_mat_prod_check_shapes = functools.partial(
    mat_prod_check_shapes, in_check=shape_like_bias, out_check=shape_like_output
)

jac_t_mat_prod_check_shapes = functools.partial(
    mat_prod_check_shapes, in_check=shape_like_output, out_check=shape_like_input
)


weight_jac_t_mat_prod_check_shapes = functools.partial(
    mat_prod_check_shapes,
    in_check=shape_like_output,
    out_check=shape_like_weight_with_sum_batch,
)
bias_jac_t_mat_prod_check_shapes = functools.partial(
    mat_prod_check_shapes,
    in_check=shape_like_output,
    out_check=shape_like_bias_with_sum_batch,
)

###############################################################################
#                     Wrapper for second-order extensions                     #
###############################################################################

# TODO Refactor using partials


def R_mat_prod_check_shapes(make_R_mat_prod):
    """Check that input and output have correct shapes."""

    @functools.wraps(make_R_mat_prod)
    def wrapped_make_R_mat_prod(self, module, g_inp, g_out):
        def checked_R_mat_prod(mat):
            check_like(mat, module, "input0")
            mat_out = make_R_mat_prod(self, module, g_inp, g_out)(mat)
            check_like(mat_out, module, "input0")
            check_same_V_dim(mat, mat_out)

            return mat_out

        return checked_R_mat_prod

    return wrapped_make_R_mat_prod


def R_mat_prod_accept_vectors(make_R_mat_prod):
    """Add support for vectors to Residual-matrix products."""

    @functools.wraps(make_R_mat_prod)
    def wrapped_make_R_mat_prod(self, module, g_inp, g_out):
        def new_R_mat_prod(mat):
            is_vec = same_dim_as(mat, module, "input0")
            mat_in = mat if not is_vec else add_V_dim(mat)
            mat_out = make_R_mat_prod(self, module, g_inp, g_out)(mat_in)
            mat_out = mat_out if not is_vec else remove_V_dim(mat_out)

            return mat_out

        return new_R_mat_prod

    return wrapped_make_R_mat_prod


def make_hessian_mat_prod_accept_vectors(make_hessian_mat_prod):
    @functools.wraps(make_hessian_mat_prod)
    def wrapped_make_hessian_mat_prod(self, module, g_inp, g_out):

        hessian_mat_prod = make_hessian_mat_prod(self, module, g_inp, g_out)

        def new_hessian_mat_prod(mat):
            is_vec = same_dim_as(mat, module, "input0")
            mat_in = mat if not is_vec else add_V_dim(mat)
            mat_out = hessian_mat_prod(mat_in)
            mat_out = mat_out if not is_vec else remove_V_dim(mat_out)

            return mat_out

        return new_hessian_mat_prod

    return wrapped_make_hessian_mat_prod


def make_hessian_mat_prod_check_shapes(make_hessian_mat_prod):
    @functools.wraps(make_hessian_mat_prod)
    def wrapped_make_hessian_mat_prod(self, module, g_inp, g_out):

        hessian_mat_prod = make_hessian_mat_prod(self, module, g_inp, g_out)

        def new_hessian_mat_prod(mat):
            check_like(mat, module, "input0")
            result = hessian_mat_prod(mat)
            check_like(result, module, "input0")

            return result

        return new_hessian_mat_prod

    return wrapped_make_hessian_mat_prod



from torch.nn import Conv2d, Linear

from backpack.extensions.backprop_extension import BackpropExtension

from . import conv2d, linear


class BatchL2Grad(BackpropExtension):
    """The squared L2 norm of individual gradients in the minibatch.

    Stores the output in ``batch_l2`` as a tensor of size ``[N]``,
    where ``N`` is the batch size.

    Note: beware of scaling issue
        The individual L2 norm depends on the scaling of the overall function.
        Let ``fᵢ`` be the loss of the ``i`` th sample, with gradient ``gᵢ``.
        ``BatchL2Grad`` will return the L2 norm of

        - ``[g₁, …, gₙ]`` if the loss is a sum, ``∑ᵢ₌₁ⁿ fᵢ``,
        - ``[¹/ₙ g₁, …, ¹/ₙ gₙ]`` if the loss is a mean, ``¹/ₙ ∑ᵢ₌₁ⁿ fᵢ``.
    """

    def __init__(self):
        super().__init__(
            savefield="batch_l2",
            fail_mode="WARNING",
            module_exts={
                Linear: linear.BatchL2Linear(),
                Conv2d: conv2d.BatchL2Conv2d(),
            },
        )



from backpack.core.derivatives.flatten import FlattenDerivatives
from backpack.extensions.secondorder.diag_ggn.diag_ggn_base import DiagGGNBaseModule


class DiagGGNFlatten(DiagGGNBaseModule):
    def __init__(self):
        super().__init__(derivatives=FlattenDerivatives())

    def backpropagate(self, ext, module, grad_inp, grad_out, backproped):
        if self.derivatives.is_no_op(module):
            return backproped
        else:
            return super().backpropagate(ext, module, grad_inp, grad_out, backproped)



from torch import einsum
from torch.nn import Unfold

from backpack.utils.ein import eingroup


def unfold_func(module):
    return Unfold(
        kernel_size=module.kernel_size,
        dilation=module.dilation,
        padding=module.padding,
        stride=module.stride,
    )


def get_weight_gradient_factors(input, grad_out, module):
    # shape [N, C_in * K_x * K_y, H_out * W_out]
    X = unfold_func(module)(input)
    dE_dY = eingroup("n,c,h,w->n,c,hw", grad_out)
    return X, dE_dY


def separate_channels_and_pixels(module, tensor):
    """Reshape (V, N, C, H, W) into (V, N, C, H * W)."""
    return eingroup("v,n,c,h,w->v,n,c,hw", tensor)


def extract_weight_diagonal(module, input, grad_output):
    """
    input must be the unfolded input to the convolution (see unfold_func)
    and grad_output the backpropagated gradient
    """
    grad_output_viewed = separate_channels_and_pixels(module, grad_output)
    AX = einsum("nkl,vnml->vnkm", (input, grad_output_viewed))
    weight_diagonal = (AX ** 2).sum([0, 1]).transpose(0, 1)
    return weight_diagonal.view_as(module.weight)


def extract_bias_diagonal(module, sqrt):
    """
    `sqrt` must be the backpropagated quantity for DiagH or DiagGGN(MC)
    """
    V_axis, N_axis = 0, 1
    bias_diagonal = (einsum("vnchw->vnc", sqrt) ** 2).sum([V_axis, N_axis])
    return bias_diagonal



"""Test implementation of `sum_hessian`"""


import pytest
import torch
from torch.nn import CrossEntropyLoss, MSELoss

from backpack import extend

from .automated_test import check_sizes, check_values
from .test_sqrt_hessian import (
    autograd_hessian,
    classification_targets,
    derivative_from_layer,
    generate_data_input_hessian,
    make_id,
    regression_targets,
)

torch.manual_seed(0)
ARGS = "layer,input_shape,targets,raises"
SETTINGS = [
    # reduction mean
    [
        # layer
        CrossEntropyLoss(reduction="mean"),
        # input_shape
        (2, 3),
        # targets
        classification_targets(shape=(2,), num_classes=3),
        # tuple of exceptions that should be raised
        (),
    ],
    [
        CrossEntropyLoss(reduction="mean"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (),
    ],
    # reduction sum
    [
        CrossEntropyLoss(reduction="sum"),
        (2, 3),
        classification_targets(shape=(2,), num_classes=3),
        (),
    ],
    [
        CrossEntropyLoss(reduction="sum"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (),
    ],
    # non-scalar outputs are not supported
    [
        CrossEntropyLoss(reduction="none"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (ValueError,),
    ],
    # no reduction for a single number as input should be fine
    [
        CrossEntropyLoss(reduction="none"),
        (1, 1),
        classification_targets(shape=(1,), num_classes=1),
        (),
    ],
    # reduction mean
    [MSELoss(reduction="mean"), (5, 1), regression_targets(shape=(5, 1)), ()],
    [MSELoss(reduction="mean"), (5, 2), regression_targets(shape=(5, 2)), ()],
    # reduction sum
    [MSELoss(reduction="sum"), (5, 1), regression_targets(shape=(5, 1)), ()],
    [MSELoss(reduction="sum"), (5, 2), regression_targets(shape=(5, 2)), ()],
    # non-scalar outputs are not supported
    [
        MSELoss(reduction="none"),
        (5, 1),
        regression_targets(shape=(5, 1)),
        (ValueError,),
    ],
]
IDS = [make_id(layer, input_shape) for (layer, input_shape, _, _) in SETTINGS]


def autograd_sum_hessian(layer, input, targets):
    """Compute the Hessian of a loss module w.r.t. its input."""
    input.requires_grad = True
    loss = layer(input, targets)
    hessian = autograd_hessian(loss, input)

    sum_hessian = sum_hessian_blocks(hessian, input)

    return sum_hessian


def sum_hessian_blocks(hessian, input):
    """Sum second derivatives over the batch dimension.
    Assert second derivative w.r.t. different samples is zero.
    """
    num_axes = len(input.shape)

    if num_axes != 2:
        raise ValueError("Only 2D inputs are currently supported.")

    N = input.shape[0]
    num_features = input.numel() // N

    sum_hessian = torch.zeros(num_features, num_features)

    hessian_different_samples = torch.zeros(num_features, num_features)
    for n_1 in range(N):
        for n_2 in range(N):
            block = hessian[n_1, :, n_2, :]

            if n_1 == n_2:
                sum_hessian += block

            else:
                assert torch.allclose(block, hessian_different_samples)

    return sum_hessian


def backpack_sum_hessian(layer, input, targets):
    layer = extend(layer)
    derivative = derivative_from_layer(layer)

    # forward pass to initialize backpack buffers
    _ = layer(input, targets)

    sum_hessian = derivative.sum_hessian(layer, None, None)
    return sum_hessian


@pytest.mark.parametrize(ARGS, SETTINGS, ids=IDS)
def test_sum_hessian(layer, input_shape, targets, raises):
    """Compare the Hessian to reconstruction from individual Hessian sqrt."""
    torch.manual_seed(0)
    input = generate_data_input_hessian(input_shape)
    try:
        return _compare_sum_hessian(layer, input, targets)
    except Exception as e:
        if isinstance(e, raises):
            return
        else:
            raise e


def _compare_sum_hessian(layer, input, targets):
    backpack_result = backpack_sum_hessian(layer, input, targets)
    autograd_result = autograd_sum_hessian(layer, input, targets)

    check_sizes(autograd_result, backpack_result)
    check_values(autograd_result, backpack_result)

    return backpack_result



from backpack.core.derivatives.conv2d import Conv2DDerivatives
from backpack.extensions.secondorder.diag_ggn.diag_ggn_base import DiagGGNBaseModule
from backpack.utils import conv as convUtils


class DiagGGNConv2d(DiagGGNBaseModule):
    def __init__(self):
        super().__init__(derivatives=Conv2DDerivatives(), params=["bias", "weight"])

    def bias(self, ext, module, grad_inp, grad_out, backproped):
        sqrt_ggn = backproped
        return convUtils.extract_bias_diagonal(module, sqrt_ggn)

    def weight(self, ext, module, grad_inp, grad_out, backproped):
        X = convUtils.unfold_func(module)(module.input0)
        weight_diag = convUtils.extract_weight_diagonal(module, X, backproped)
        return weight_diag



from backpack.extensions.firstorder.base import FirstOrderModuleExtension


class BatchGradBase(FirstOrderModuleExtension):
    def __init__(self, derivatives, params=None):
        self.derivatives = derivatives
        super().__init__(params=params)

    def bias(self, ext, module, g_inp, g_out, bpQuantities):
        return self.derivatives.bias_jac_t_mat_prod(
            module, g_inp, g_out, g_out[0], sum_batch=False
        )

    def weight(self, ext, module, g_inp, g_out, bpQuantities):
        return self.derivatives.weight_jac_t_mat_prod(
            module, g_inp, g_out, g_out[0], sum_batch=False
        )



from backpack.core.derivatives.avgpool2d import AvgPool2DDerivatives
from backpack.core.derivatives.maxpool2d import MaxPool2DDerivatives
from backpack.extensions.secondorder.diag_ggn.diag_ggn_base import DiagGGNBaseModule


class DiagGGNMaxPool2d(DiagGGNBaseModule):
    def __init__(self):
        super().__init__(derivatives=MaxPool2DDerivatives())


class DiagGGNAvgPool2d(DiagGGNBaseModule):
    def __init__(self):
        super().__init__(derivatives=AvgPool2DDerivatives())



from backpack.core.derivatives.avgpool2d import AvgPool2DDerivatives
from backpack.core.derivatives.maxpool2d import MaxPool2DDerivatives

from .cmpbase import CMPBase


class CMPAvgPool2d(CMPBase, AvgPool2DDerivatives):
    def __init__(self):
        super().__init__(derivatives=AvgPool2DDerivatives())


class CMPMaxpool2d(CMPBase, MaxPool2DDerivatives):
    def __init__(self):
        super().__init__(derivatives=MaxPool2DDerivatives())



from torch.nn import Conv2d, Linear

from backpack.extensions.backprop_extension import BackpropExtension

from . import conv2d, linear


class Variance(BackpropExtension):
    """Estimates the variance of the gradient using the samples in the minibatch.

    Stores the output in ``variance``. Same dimension as the gradient.

    Note: beware of scaling issue
        The variance depends on the scaling of the overall function.
        Let ``fᵢ`` be the loss of the ``i`` th sample, with gradient ``gᵢ``.
        ``Variance`` will return the variance of the vectors

        - ``[g₁, …, gₙ]`` if the loss is a sum, ``∑ᵢ₌₁ⁿ fᵢ``,
        - ``[¹/ₙ g₁, …, ¹/ₙ gₙ]`` if the loss is a mean, ``¹/ₙ ∑ᵢ₌₁ⁿ fᵢ``.
    """

    def __init__(self):
        super().__init__(
            savefield="variance",
            fail_mode="WARNING",
            module_exts={
                Linear: linear.VarianceLinear(),
                Conv2d: conv2d.VarianceConv2d(),
            },
        )



"""Second order extensions.
====================================

Second-order extensions propagate additional information through the graph
to extract structural or local approximations to second-order information.
They are more expensive to run than a standard gradient backpropagation.
The implemented extensions are

- The diagonal of the Generalized Gauss-Newton (GGN)/Fisher information,
  using exact computation
  (:func:`DiagGGNExact <backpack.extensions.DiagGGNExact>`)
  or Monte-Carlo approximation
  (:func:`DiagGGNMC <backpack.extensions.DiagGGNMC>`).
- Kronecker Block-Diagonal approximations of the GGN/Fisher
  :func:`KFAC <backpack.extensions.KFAC>`,
  :func:`KFRA <backpack.extensions.KFRA>`,
  :func:`KFLR <backpack.extensions.KFLR>`.
- The diagonal of the Hessian :func:`DiagHessian <backpack.extensions.DiagHessian>`
"""

from .diag_ggn import DiagGGNExact, DiagGGNMC, DiagGGN
from .diag_hessian import DiagHessian
from .hbp import KFAC, KFLR, KFRA, HBP

__all__ = [
    "DiagGGNExact",
    "DiagGGNMC",
    "DiagGGN",
    "DiagHessian",
    "KFAC",
    "KFLR",
    "KFRA",
    "HBP",
]



"""Checks wether backpack correctly recognizes when 2nd order backprop would fail.

Failure cases include
- the loss if not the output of a torch.nn module
- using unsupported parameters of the loss
"""

import pytest
import torch
from torch.nn import CrossEntropyLoss, MSELoss
from backpack import extend
from backpack import backpack as bp
import backpack.extensions as bpext

ext_2nd_order = [
    bpext.KFAC,
    bpext.KFRA,
    bpext.KFLR,
    bpext.DiagGGNExact,
    bpext.DiagGGNMC,
]
ext_2nd_order_name = [
    "KFAC",
    "KFRA",
    "KFLR",
    "DiagGGNExact",
    "DiagGGNExactMC",
]


def classification_targets(N, num_classes):
    """Create random targets for classes 0, ..., `num_classes - 1`."""
    return torch.randint(size=(N,), low=0, high=num_classes)


def dummy_cross_entropy(N=5):
    y_pred = torch.rand((N, 2))
    y_pred.requires_grad = True
    y = classification_targets(N, 2)
    loss_module = extend(CrossEntropyLoss())
    return loss_module(y_pred, y)


def dummy_mse(N=5, D=1):
    y_pred = torch.rand((N, D))
    y_pred.requires_grad = True
    y = torch.randn((N, D))
    loss_module = extend(MSELoss())
    return loss_module(y_pred, y)


@pytest.mark.parametrize("extension", ext_2nd_order, ids=ext_2nd_order_name)
def test_sqrt_hessian_crossentropy_should_pass(extension):
    loss = dummy_cross_entropy()

    with bp(extension()):
        loss.backward()


@pytest.mark.parametrize("extension", ext_2nd_order, ids=ext_2nd_order_name)
def test_sqrt_hessian_mse_should_pass(extension):
    loss = dummy_mse()

    with bp(extension()):
        loss.backward()


@pytest.mark.parametrize("extension", ext_2nd_order, ids=ext_2nd_order_name)
def test_sqrt_hessian_modified_crossentropy_should_fail(extension):
    loss = dummy_cross_entropy() * 2

    with pytest.warns(UserWarning):
        with bp(extension()):
            loss.backward()


@pytest.mark.parametrize("extension", ext_2nd_order, ids=ext_2nd_order_name)
def test_sqrt_hessian_modified_mse_should_fail(extension):
    loss = dummy_mse() * 2

    with pytest.warns(UserWarning):
        with bp(extension()):
            loss.backward()



from torch import einsum

from backpack.extensions.firstorder.base import FirstOrderModuleExtension


class SGSLinear(FirstOrderModuleExtension):
    def __init__(self):
        super().__init__(params=["bias", "weight"])

    def bias(self, ext, module, g_inp, g_out, backproped):
        N_axis = 0
        return (g_out[0] ** 2).sum(N_axis)

    def weight(self, ext, module, g_inp, g_out, backproped):
        return einsum("ni,nj->ij", (g_out[0] ** 2, module.input0 ** 2))



from torch.nn import (
    AvgPool2d,
    Conv2d,
    CrossEntropyLoss,
    MSELoss,
    Dropout,
    Linear,
    MaxPool2d,
    ReLU,
    Sigmoid,
    Tanh,
    ZeroPad2d,
)

from .avgpool2d import AvgPool2DDerivatives
from .conv2d import Conv2DDerivatives
from .crossentropyloss import CrossEntropyLossDerivatives
from .mseloss import MSELossDerivatives
from .dropout import DropoutDerivatives
from .linear import LinearDerivatives
from .maxpool2d import MaxPool2DDerivatives
from .relu import ReLUDerivatives
from .sigmoid import SigmoidDerivatives
from .tanh import TanhDerivatives
from .zeropad2d import ZeroPad2dDerivatives

derivatives_for = {
    Linear: LinearDerivatives,
    Conv2d: Conv2DDerivatives,
    AvgPool2d: AvgPool2DDerivatives,
    MaxPool2d: MaxPool2DDerivatives,
    ZeroPad2d: ZeroPad2dDerivatives,
    Dropout: DropoutDerivatives,
    ReLU: ReLUDerivatives,
    Tanh: TanhDerivatives,
    Sigmoid: SigmoidDerivatives,
    CrossEntropyLoss: CrossEntropyLossDerivatives,
    MSELoss: MSELossDerivatives,
}



from torch.nn import Sigmoid

from backpack.core.derivatives.elementwise import ElementwiseDerivatives


class SigmoidDerivatives(ElementwiseDerivatives):
    def get_module(self):
        return Sigmoid

    def hessian_is_zero(self):
        return False

    def hessian_is_diagonal(self):
        return True

    def df(self, module, g_inp, g_out):
        return module.output * (1.0 - module.output)

    def d2f(self, module, g_inp, g_out):
        return module.output * (1 - module.output) * (1 - 2 * module.output)



from backpack.extensions.firstorder.base import FirstOrderModuleExtension


class VarianceBaseModule(FirstOrderModuleExtension):
    def __init__(self, params, grad_extension, sgs_extension):
        super().__init__(params=params)
        self.grad_ext = grad_extension
        self.sgs_ext = sgs_extension

    @staticmethod
    def variance_from(grad, sgs, N):
        avgg_squared = (grad / N) ** 2
        avg_gsquared = sgs / N
        return avg_gsquared - avgg_squared

    def bias(self, ext, module, g_inp, g_out, backproped):
        N = g_out[0].shape[0]
        return self.variance_from(
            self.grad_ext.bias(ext, module, g_inp, g_out, backproped),
            self.sgs_ext.bias(ext, module, g_inp, g_out, backproped),
            N,
        )

    def weight(self, ext, module, g_inp, g_out, backproped):
        N = g_out[0].shape[0]
        return self.variance_from(
            self.grad_ext.weight(ext, module, g_inp, g_out, backproped),
            self.sgs_ext.weight(ext, module, g_inp, g_out, backproped),
            N,
        )



from torch.nn import BatchNorm1d, Conv2d, Linear

from backpack.extensions.backprop_extension import BackpropExtension

from . import batchnorm1d, conv2d, linear


class BatchGrad(BackpropExtension):
    """Individual gradients for each sample in a minibatch.

    Stores the output in ``grad_batch`` as a ``[N x ...]`` tensor,
    where ``N`` batch size and ``...`` is the shape of the gradient.

    Note: beware of scaling issue
        The `individual gradients` depend on the scaling of the overall function.
        Let ``fᵢ`` be the loss of the ``i`` th sample, with gradient ``gᵢ``.
        ``BatchGrad`` will return

        - ``[g₁, …, gₙ]`` if the loss is a sum, ``∑ᵢ₌₁ⁿ fᵢ``,
        - ``[¹/ₙ g₁, …, ¹/ₙ gₙ]`` if the loss is a mean, ``¹/ₙ ∑ᵢ₌₁ⁿ fᵢ``.

    The concept of individual gradients is only meaningful if the
    objective is a sum of independent functions (no batchnorm).

    """

    def __init__(self):
        super().__init__(
            savefield="grad_batch",
            fail_mode="WARNING",
            module_exts={
                Linear: linear.BatchGradLinear(),
                Conv2d: conv2d.BatchGradConv2d(),
                BatchNorm1d: batchnorm1d.BatchGradBatchNorm1d(),
            },
        )



"""BackPACK."""
import inspect

import torch
from backpack.extensions.backprop_extension import BackpropExtension

from . import extensions
from .context import CTX


class backpack:
    """Activates Backpack Extensions.

    Activates the BackPACK extensions passed as arguments for the
    :code:`backward` calls in the current :code:`with` block.
    """

    def __init__(self, *exts: BackpropExtension, debug=False):
        """Activate the Backpack extensions.

        Example usage:
        ```
        X, Y, model, lossfunc = get_problem()

        backpack.extend(model)
        backpack.extend(lossfunc)

        with backpack.backpack(backpack.extensions.Variance()):
            lossfunc(model(X), Y).backward()

            for p in model.parameters():
                print(p.grad)
                print(p.variance)
        ```

        .. warning ::

            The quantities computed by backPACK may be garbage collected when
            exiting the `with` clause. Use them within the `with` clause or
            assign them to another variable.

        Attributes:
            args: [BackpropExtension]
                The extensions to activate for the backward pass.
            debug: Bool, optional (default: False)
                If true, will print debug messages during the backward pass.
        """
        for ext in exts:
            if not isinstance(ext, BackpropExtension):
                if inspect.isclass(ext) and issubclass(ext, BackpropExtension):
                    raise ValueError(
                        "backpack expect instances of BackpropExtension,"
                        + " but received a class instead [{}].".format(ext)
                        + " Instantiate it before passing it to backpack."
                    )
                else:
                    raise ValueError(
                        "backpack expects instances of BackpropExtension,"
                        + " but received [{}].".format(ext)
                    )

        self.exts = exts
        self.debug = debug

    def __enter__(self):
        self.old_CTX = CTX.get_active_exts()
        self.old_debug = CTX.get_debug()
        CTX.set_active_exts(self.exts)
        CTX.set_debug(self.debug)

    def __exit__(self, type, value, traceback):
        CTX.set_active_exts(self.old_CTX)
        CTX.set_debug(self.old_debug)


def hook_store_io(module, input, output):
    """Saves the input and output as attributes of the module.

    Args:
        module: module
        input: List of input tensors
        output: output tensor
    """
    for i in range(len(input)):
        setattr(module, "input{}".format(i), input[i])
    module.output = output


def hook_store_shapes(module, input, output):
    """Store dimensionality of output as buffer.

    Args:
        module: module
        input: List of input tensors shapes
        output: output tensor shape
    """
    for i in range(len(input)):
        module.register_buffer(
            "input{}_shape".format(i), torch.IntTensor([*input[i].size()])
        )
    module.register_buffer("output_shape", torch.IntTensor([*output.size()]))


def memory_cleanup(module):
    """Remove I/O stored by backpack during the forward pass.

    Deletes the attributes created by `hook_store_io` and `hook_store_shapes`.
    """
    if hasattr(module, "output"):
        delattr(module, "output")
    if hasattr(module, "output_shape"):
        delattr(module, "output_shape")
    i = 0
    while hasattr(module, "input{}".format(i)):
        delattr(module, "input{}".format(i))
        i += 1
    i = 0
    while hasattr(module, "input{}_shape".format(i)):
        delattr(module, "input{}_shape".format(i))
        i += 1


def hook_run_extensions(module, g_inp, g_out):
    for backpack_extension in CTX.get_active_exts():
        if CTX.get_debug():
            print("[DEBUG] Running extension", backpack_extension, "on", module)
        backpack_extension.apply(module, g_inp, g_out)

    if not CTX.is_extension_active(extensions.curvmatprod.CMP):
        memory_cleanup(module)


def extend(module: torch.nn.Module, debug=False):
    """Extends the ``module`` to make it backPACK-ready.

    If the ``module`` has children, e.g. for a ``torch.nn.Sequential``,
    they will also be extended.

    Args:
        module: torch.nn.Module
            The module to extend
        debug: Bool, optional (default: False)
            If true, will print debug messages during the extension.
    """
    if debug:
        print("[DEBUG] Extending", module)

    for child in module.children():
        extend(child, debug=debug)

    module_was_already_extended = getattr(module, "_backpack_extend", False)
    if not module_was_already_extended:
        CTX.add_hook_handle(module.register_forward_hook(hook_store_io))
        CTX.add_hook_handle(module.register_forward_hook(hook_store_shapes))
        CTX.add_hook_handle(module.register_backward_hook(hook_run_extensions))
        module._backpack_extend = True

    return module



from backpack.core.derivatives.relu import ReLUDerivatives
from backpack.core.derivatives.sigmoid import SigmoidDerivatives
from backpack.core.derivatives.tanh import TanhDerivatives
from backpack.extensions.secondorder.diag_ggn.diag_ggn_base import DiagGGNBaseModule


class DiagGGNReLU(DiagGGNBaseModule):
    def __init__(self):
        super().__init__(derivatives=ReLUDerivatives())


class DiagGGNSigmoid(DiagGGNBaseModule):
    def __init__(self):
        super().__init__(derivatives=SigmoidDerivatives())


class DiagGGNTanh(DiagGGNBaseModule):
    def __init__(self):
        super().__init__(derivatives=TanhDerivatives())



""" TODO: Implement all features for BN, then add to automated tests. """
import pytest
import torch

from .automated_test import check_sizes, check_values
from .implementation.implementation_autograd import AutogradImpl
from .implementation.implementation_bpext import BpextImpl
from .test_problems_bn import TEST_PROBLEMS as BN_TEST_PROBLEMS

if torch.cuda.is_available():
    DEVICES = {
        "cpu": "cpu",
        "gpu": "cuda:0",
    }
else:
    DEVICES = {
        "cpu": "cpu",
    }

TEST_PROBLEMS = {
    **BN_TEST_PROBLEMS,
}

ALL_CONFIGURATIONS = []
CONFIGURATION_IDS = []
for dev_name, dev in DEVICES.items():
    for probname, prob in TEST_PROBLEMS.items():
        ALL_CONFIGURATIONS.append((prob, dev))
        CONFIGURATION_IDS.append(probname + "-" + dev_name)


###
# Tests
###
@pytest.mark.parametrize("problem,device", ALL_CONFIGURATIONS, ids=CONFIGURATION_IDS)
def test_batch_gradients_sum_to_grad(problem, device):
    problem.to(device)
    backpack_batch_res = BpextImpl(problem).batch_gradients()
    backpack_res = [g.sum(0) for g in backpack_batch_res]
    autograd_res = AutogradImpl(problem).gradient()

    check_sizes(autograd_res, backpack_res, list(problem.model.parameters()))
    check_values(autograd_res, backpack_res)


@pytest.mark.parametrize("problem,device", ALL_CONFIGURATIONS, ids=CONFIGURATION_IDS)
def test_ggn_mp(problem, device):
    problem.to(device)

    NUM_COLS = 10
    matrices = [
        torch.randn(NUM_COLS, *p.shape, device=device)
        for p in problem.model.parameters()
    ]

    autograd_res = AutogradImpl(problem).ggn_mp(matrices)
    backpack_res = BpextImpl(problem).ggn_mp(matrices)

    check_sizes(autograd_res, backpack_res)
    check_values(autograd_res, backpack_res)


@pytest.mark.parametrize("problem,device", ALL_CONFIGURATIONS, ids=CONFIGURATION_IDS)
def test_ggn_vp(problem, device):
    problem.to(device)

    vecs = [torch.randn(*p.shape, device=device) for p in problem.model.parameters()]

    backpack_res = BpextImpl(problem).ggn_vp(vecs)
    autograd_res = AutogradImpl(problem).ggn_vp(vecs)

    check_sizes(autograd_res, backpack_res)
    check_values(autograd_res, backpack_res)


@pytest.mark.parametrize("problem,device", ALL_CONFIGURATIONS, ids=CONFIGURATION_IDS)
def test_hvp_is_not_implemented(problem, device):
    # TODO: Rename after implementing BatchNorm R_mat_prod
    problem.to(device)

    vecs = [torch.randn(*p.shape, device=device) for p in problem.model.parameters()]

    # TODO: Implement BatchNorm R_mat_prod in backpack/core/derivatives/batchnorm1d.py
    try:
        backpack_res = BpextImpl(problem).hvp(vecs)
    except NotImplementedError:
        return

    autograd_res = AutogradImpl(problem).hvp(vecs)

    check_sizes(autograd_res, backpack_res)
    check_values(autograd_res, backpack_res)


@pytest.mark.parametrize("problem,device", ALL_CONFIGURATIONS, ids=CONFIGURATION_IDS)
def test_hmp_is_not_implemented(problem, device):
    # TODO: Rename after implementing BatchNorm R_mat_prod
    problem.to(device)

    NUM_COLS = 10
    matrices = [
        torch.randn(NUM_COLS, *p.shape, device=device)
        for p in problem.model.parameters()
    ]

    # TODO: Implement BatchNorm R_mat_prod in backpack/core/derivatives/batchnorm1d.py
    try:
        backpack_res = BpextImpl(problem).hmp(matrices)
    except NotImplementedError:
        return

    autograd_res = AutogradImpl(problem).hmp(matrices)

    check_sizes(autograd_res, backpack_res)
    check_values(autograd_res, backpack_res)



from torch.nn import ZeroPad2d, functional

from backpack.core.derivatives.basederivatives import BaseDerivatives
from backpack.utils.ein import eingroup


class ZeroPad2dDerivatives(BaseDerivatives):
    def get_module(self):
        return ZeroPad2d

    def hessian_is_zero(self):
        return True

    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        _, C_out, H_out, W_out = module.output_shape
        _, in_c, in_x, in_y = module.input0_shape
        in_features = in_c * in_x * in_y

        result = mat.view(C_out, H_out, W_out, C_out, H_out, W_out)

        (W_top, W_bottom), (H_bottom, H_top) = self.__unpad_indices(module)
        result = result[
            :, W_top:W_bottom, H_bottom:H_top, :, W_top:W_bottom, H_bottom:H_top,
        ].contiguous()

        return result.view(in_features, in_features)

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        (W_top, W_bottom), (H_bottom, H_top) = self.__unpad_indices(module)
        return mat[:, :, :, W_top:W_bottom, H_bottom:H_top]

    def __unpad_indices(self, module):
        _, _, H_out, W_out = module.output_shape
        pad_left, pad_right, pad_top, pad_bottom = module.padding

        H_bottom, H_top = pad_left, W_out - pad_right
        W_top, W_bottom = pad_top, H_out - pad_bottom

        return (W_top, W_bottom), (H_bottom, H_top)

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        mat = eingroup("v,n,c,h,w->vn,c,h,w", mat)
        pad_mat = functional.pad(mat, module.padding, "constant", module.value)
        return self.reshape_like_output(pad_mat, module)



from backpack.core.derivatives.avgpool2d import AvgPool2DDerivatives
from backpack.core.derivatives.maxpool2d import MaxPool2DDerivatives
from backpack.extensions.secondorder.diag_hessian.diag_h_base import DiagHBaseModule


class DiagHAvgPool2d(DiagHBaseModule):
    def __init__(self):
        super().__init__(derivatives=AvgPool2DDerivatives())


class DiagHMaxPool2d(DiagHBaseModule):
    def __init__(self):
        super().__init__(derivatives=MaxPool2DDerivatives())



"""Base classes for more flexible Jacobians and second-order information."""

import warnings

from backpack.core.derivatives import shape_check


class BaseDerivatives:
    """First- and second-order partial derivatives of unparameterized module.

    Note:
        Throughout the code, use these conventions if possible:

        - `N`: batch size
        - Vectors
          - Layer input shape `[N, D_in]`
          - Layer output shape `[N, D_out]`
        - Images
          - Layer input shape `[N, C_in, H_in, W_in]`
          - Layer output shape `[N, C_out, H_out, W_out]`
        - `V`: vectorization axis

    Definition:
        For simplicity, consider the vector case, i.e. a function which maps an
        `[N, D_in]` `input` into an `[N, D_out]` `output`.

        The Jacobian `J` of  is tensor of shape `[N, D_out, N_in, D_in]`.
        Partial derivatives are ordered as

            `J[i, j, k, l] = 𝜕output[i, j] / 𝜕input[k, l].

        The transposed Jacobian `Jᵀ` has shape `[N, D_in, N, D_out]`.
        Partial derivatives are ordered as

            `Jᵀ[i, j, k, l] = 𝜕output[k, l] / 𝜕input[i, j]`.

        In general, feature dimension indices `j, l` are product indices.
    """

    @shape_check.jac_mat_prod_accept_vectors
    @shape_check.jac_mat_prod_check_shapes
    def jac_mat_prod(self, module, g_inp, g_out, mat):
        """Apply Jacobian of the output w.r.t. input to a matrix.

        It is assumed that the module input has shape `[N, *]`, while the output is
        of shape `[N, •]`. Both `*`, `•` denote arbitrary shapes.

        Apply Jacobian to all slices among the vectorization axis.

            `result[v, n, •] =  ∑ₖ ∑_* J[n, •, k, *] mat[v, n, *]`.

        Args:
            module (torch.nn.Module): Extended module.
            g_inp ([torch.Tensor]): Gradients of the module w.r.t. its inputs.
            g_out ([torch.Tensor]): Gradients of the module w.r.t. its outputs.
            mat (torch.Tensor): Matrix the Jacobian will be applied to. Must have
                shape `[V, N, *]`.

        Returns:
            torch.Tensor: Jacobian-matrix product. Has shape [V, N, *].

        Note:
            - The Jacobian can be applied without knowledge about backpropagated
              derivatives. Both `g_inp` and `g_out` are usually not required and
              can be set to `None`.
        """
        return self._jac_mat_prod(module, g_inp, g_out, mat)

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        """Internal implementation of the Jacobian."""
        raise NotImplementedError

    @shape_check.jac_t_mat_prod_accept_vectors
    @shape_check.jac_t_mat_prod_check_shapes
    def jac_t_mat_prod(self, module, g_inp, g_out, mat):
        """Apply transposed Jacobian of module output w.r.t. input to a matrix.

        Implicit application of Jᵀ:
            result[v, ̃n, ̃c, ̃w, ...]
            = ∑_{n, c, w} Jᵀ[̃n, ̃c, ̃w, ..., n, c, w, ...] mat[v, n, c, w, ...].

        Parameters:
        -----------
        mat: torch.Tensor
            Matrix the transposed Jacobian will be applied to.
            Must have shape [V, N, C_out, H_out, ...].

        Returns:
        --------
        result: torch.Tensor
            Transposed Jacobian-matrix product.
            Has shape [V, N, C_in, H_in, ...].
        """
        return self._jac_t_mat_prod(module, g_inp, g_out, mat)

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        """Internal implementation of transposed Jacobian."""
        raise NotImplementedError

    # TODO Add shape check
    # TODO Use new convention
    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        raise NotImplementedError

    def hessian_is_zero(self):
        raise NotImplementedError

    def hessian_is_diagonal(self):
        raise NotImplementedError

    def hessian_diagonal(self):
        raise NotImplementedError

    def hessian_is_psd(self):
        raise NotImplementedError

    # TODO make accept vectors
    # TODO add shape check
    def make_residual_mat_prod(self, module, g_inp, g_out):
        """Return multiplication routine with the residual term.

        The function performs the mapping: mat → [∑_{k} Hz_k(x) 𝛿z_k] mat.
        (required for extension `curvmatprod`)

        Note:
        -----
            This function only has to be implemented if the residual is not
            zero and not diagonal (for instance, `BatchNorm`).
        """
        raise NotImplementedError

    # TODO Refactor and remove
    def batch_flat(self, tensor):
        batch = tensor.size(0)
        # TODO Removing the clone().detach() will destroy the computation graph
        # Tests will fail
        return batch, tensor.clone().detach().view(batch, -1)

    # TODO Refactor and remove
    def get_batch(self, module):
        return module.input0.size(0)

    @staticmethod
    def _reshape_like(mat, like):
        """Reshape as like with trailing and additional 0th dimension.

        If like is [N, C, H, ...], returns shape [-1, N, C, H, ...]
        """
        V = -1
        shape = (V, *like.shape)
        return mat.reshape(shape)

    @classmethod
    def reshape_like_input(cls, mat, module):
        return cls._reshape_like(mat, module.input0)

    @classmethod
    def reshape_like_output(cls, mat, module):
        return cls._reshape_like(mat, module.output)


class BaseParameterDerivatives(BaseDerivatives):
    """First- and second order partial derivatives of a module with parameters.

    Assumptions (true for `nn.Linear`, `nn.Conv(Transpose)Nd`, `nn.BatchNormNd`):
    - Parameters are saved as `.weight` and `.bias` fields in a module
    - The output is linear in the model parameters

    Shape conventions:
    ------------------
    Weight [C_w, H_w, W_w, ...] (usually 1d, 2d, 4d)
    Bias [C_b, ...] (usually 1d)

    For most layers, these shapes correspond to shapes of the module input or output.
    """

    @shape_check.bias_jac_mat_prod_accept_vectors
    @shape_check.bias_jac_mat_prod_check_shapes
    def bias_jac_mat_prod(self, module, g_inp, g_out, mat):
        """Apply Jacobian of the output w.r.t. bias to a matrix.

        Parameters:
        -----------
        mat: torch.Tensor
            Matrix the Jacobian will be applied to.
            Must have shape [V, C_b, ...].

        Returns:
        --------
        result: torch.Tensor
            Jacobian-matrix product.
            Has shape [V, N, C_out, H_out, ...].
        """
        return self._bias_jac_mat_prod(module, g_inp, g_out, mat)

    def _bias_jac_mat_prod(self, module, g_inp, g_out, mat):
        """Internal implementation of the bias Jacobian."""
        raise NotImplementedError

    @shape_check.bias_jac_t_mat_prod_accept_vectors
    @shape_check.bias_jac_t_mat_prod_check_shapes
    def bias_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        """Apply transposed Jacobian of the output w.r.t. bias to a matrix.

        Parameters:
        -----------
        mat: torch.Tensor
            Matrix the transposed Jacobian will be applied to.
            Must have shape [V, N, C_out, H_out, ...].
        sum_batch: bool
            Whether to sum over the batch dimension on the fly.

        Returns:
        --------
        result: torch.Tensor
            Jacobian-matrix product.
            Has shape [V, N, C_b, ...] if `sum_batch == False`.
            Has shape [V, C_b, ...] if `sum_batch == True`.
        """
        return self._bias_jac_t_mat_prod(module, g_inp, g_out, mat, sum_batch=sum_batch)

    def _bias_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        """Internal implementation of the transposed bias Jacobian."""
        raise NotImplementedError

    @shape_check.weight_jac_mat_prod_accept_vectors
    @shape_check.weight_jac_mat_prod_check_shapes
    def weight_jac_mat_prod(self, module, g_inp, g_out, mat):
        """Apply Jacobian of the output w.r.t. weight to a matrix.

        Parameters:
        -----------
        mat: torch.Tensor
            Matrix the Jacobian will be applied to.
            Must have shape [V, C_w, H_w, ...].

        Returns:
        --------
        result: torch.Tensor
            Jacobian-matrix product.
            Has shape [V, N, C_out, H_out, ...].
        """
        return self._weight_jac_mat_prod(module, g_inp, g_out, mat)

    def _weight_jac_mat_prod(self, module, g_inp, g_out, mat):
        """Internal implementation of weight Jacobian."""
        raise NotImplementedError

    @shape_check.weight_jac_t_mat_prod_accept_vectors
    @shape_check.weight_jac_t_mat_prod_check_shapes
    def weight_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        """Apply transposed Jacobian of the output w.r.t. weight to a matrix.

        Parameters:
        -----------
        mat: torch.Tensor
            Matrix the transposed Jacobian will be applied to.
            Must have shape [V, N, C_out, H_out, ...].
        sum_batch: bool
            Whether to sum over the batch dimension on the fly.

        Returns:
        --------
        result: torch.Tensor
            Jacobian-matrix product.
            Has shape [V, N, C_w, H_w, ...] if `sum_batch == False`.
            Has shape [V, C_w, H_w, ...] if `sum_batch == True`.
        """
        return self._weight_jac_t_mat_prod(
            module, g_inp, g_out, mat, sum_batch=sum_batch
        )

    def _weight_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        """Internal implementation of transposed weight Jacobian."""
        raise NotImplementedError


class BaseLossDerivatives(BaseDerivatives):
    """Second- order partial derivatives of loss functions."""

    # TODO Add shape check
    def sqrt_hessian(self, module, g_inp, g_out):
        """Symmetric factorization ('sqrt') of the loss Hessian."""
        self.check_2nd_order_make_sense(module, g_inp, g_out)
        return self._sqrt_hessian(module, g_inp, g_out)

    def _sqrt_hessian(self, module, g_inp, g_out):
        raise NotImplementedError

    # TODO Add shape check
    def sqrt_hessian_sampled(self, module, g_inp, g_out, mc_samples=1):
        """Monte-Carlo sampled symmetric factorization of the loss Hessian."""
        self.check_2nd_order_make_sense(module, g_inp, g_out)
        return self._sqrt_hessian_sampled(module, g_inp, g_out, mc_samples=mc_samples)

    def _sqrt_hessian_sampled(self, module, g_inp, g_out, mc_samples=1):
        raise NotImplementedError

    @shape_check.make_hessian_mat_prod_accept_vectors
    @shape_check.make_hessian_mat_prod_check_shapes
    def make_hessian_mat_prod(self, module, g_inp, g_out):
        """Multiplication of the input Hessian with a matrix.

        Return a function that maps mat to H * mat.
        """
        self.check_2nd_order_make_sense(module, g_inp, g_out)
        return self._make_hessian_mat_prod(module, g_inp, g_out)

    def _make_hessian_mat_prod(self, module, g_inp, g_out):
        raise NotImplementedError

    # TODO Add shape check
    def sum_hessian(self, module, g_inp, g_out):
        """Loss Hessians, summed over the batch dimension."""
        self.check_2nd_order_make_sense(module, g_inp, g_out)
        return self._sum_hessian(module, g_inp, g_out)

    def _sum_hessian(self, module, g_inp, g_out):
        raise NotImplementedError

    def check_2nd_order_make_sense(self, module, g_inp, g_out):
        """Verify conditions for 2nd-order extensions to be working.

        2nd-order extensions are only guaranteed to work if the `loss`,
        on which `backward()` is called, is a scalar that has not been
        modified further after passing through the loss function module.
        """
        self._check_output_is_scalar(module)
        self._check_loss_has_not_been_modified(module, g_out)

    def _check_output_is_scalar(self, module):
        """Raise an exception is the module output is not a scalar."""
        if module.output.numel() != 1:
            raise ValueError(
                "Output must be scalar. Got {}".format(module.output.shape)
            )

    def _check_loss_has_not_been_modified(self, module, g_out):
        """Raise a warning if the module output seems to have been changed."""
        grad_out_is_identity = g_out is None or (g_out[0] == 1.0).all().item()
        if not grad_out_is_identity:
            warnings.warn(
                "The output of {} seems to have been modified.".format(module)
                + " Backpack might give wrong second-order information."
                + " Make sure you call backward() on the output of a loss"
                + " function module from torch.nn",
                UserWarning,
            )



from torch.nn import Tanh

from backpack.core.derivatives.elementwise import ElementwiseDerivatives


class TanhDerivatives(ElementwiseDerivatives):
    def get_module(self):
        return Tanh

    def hessian_is_zero(self):
        return False

    def hessian_is_diagonal(self):
        return True

    def df(self, module, g_inp, g_out):
        return 1.0 - module.output ** 2

    def d2f(self, module, g_inp, g_out):
        return -2.0 * module.output * (1.0 - module.output ** 2)



from torch import einsum
from torch.nn import Linear

from backpack.core.derivatives.basederivatives import BaseParameterDerivatives


class LinearDerivatives(BaseParameterDerivatives):
    """Partial derivatives for the Linear layer.

    Index conventions:
    ------------------
    * v: Free dimension
    * n: Batch dimension
    * o: Output dimension
    * i: Input dimension
    """

    def get_module(self):
        return Linear

    def hessian_is_zero(self):
        return True

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        """Apply transposed Jacobian of the output w.r.t. the input."""
        d_input = module.weight.data
        return einsum("oi,vno->vni", (d_input, mat))

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        """Apply Jacobian of the output w.r.t. the input."""
        d_input = module.weight.data
        return einsum("oi,vni->vno", (d_input, mat))

    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        jac = module.weight.data
        return einsum("ik,ij,jl->kl", (jac, mat, jac))

    def _weight_jac_mat_prod(self, module, g_inp, g_out, mat):
        """Apply Jacobian of the output w.r.t. the weight."""
        d_weight = module.input0
        return einsum("ni,voi->vno", (d_weight, mat))

    def _weight_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        """Apply transposed Jacobian of the output w.r.t. the weight."""
        d_weight = module.input0
        contract = "vno,ni->voi" if sum_batch else "vno,ni->vnoi"
        return einsum(contract, (mat, d_weight))

    def _bias_jac_mat_prod(self, module, g_inp, g_out, mat):
        """Apply Jacobian of the output w.r.t. the bias."""
        N = self.get_batch(module)
        return mat.unsqueeze(1).expand(-1, N, -1)

    def _bias_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        """Apply transposed Jacobian of the output w.r.t. the bias."""
        if sum_batch:
            N_axis = 1
            return mat.sum(N_axis)
        else:
            return mat



"""Derivatives of the MSE Loss."""

from math import sqrt

from torch import eye, einsum, normal
from torch.nn import MSELoss

from backpack.core.derivatives.basederivatives import BaseLossDerivatives


class MSELossDerivatives(BaseLossDerivatives):
    """Derivatives of the MSE Loss.

    We only support 2D tensors.

    For `X : [n, d]` and `Y : [n, d]`, if `reduce=sum`, the MSE computes
    `∑ᵢ₌₁ⁿ ‖X[i,∶] − Y[i,∶]‖²`. If `reduce=mean`, the result is divided by `nd`.


    """

    def get_module(self):
        return MSELoss

    def _sqrt_hessian(self, module, g_inp, g_out):
        """Square-root of the hessian of the MSE for each minibatch elements.

        Returns the Hessian in format `Hs = [D, N, D]`, where
        `Hs[:, n, :]` is the Hessian for the `n`th element.

        Attributes:
            module: (torch.nn.MSELoss) module
            g_inp: Gradient of loss w.r.t. input
            g_out: Gradient of loss w.r.t. output

        Returns:
             Batch of hessians, in format [D, N, D]
        """
        self.check_input_dims(module)

        N, D = module.input0.shape
        sqrt_H = sqrt(2) * eye(D, device=module.input0.device)  # [D, D]
        sqrt_H = sqrt_H.unsqueeze(0).repeat(N, 1, 1)  # [N, D, D]
        sqrt_H = einsum("nab->anb", sqrt_H)  # [D, N, D]

        if module.reduction == "mean":
            sqrt_H /= sqrt(module.input0.numel())

        return sqrt_H

    def _sqrt_hessian_sampled(self, module, g_inp, g_out, mc_samples=1):
        """A Monte-Carlo estimate of the square-root of the Hessian.

        Attributes:
            module: (torch.nn.MSELoss) module.
            g_inp: Gradient of loss w.r.t. input.
            g_out: Gradient of loss w.r.t. output.
            mc_samples: (int, optional) Number of MC samples to use. Default: 1.

        Returns:
            tensor:
        """
        N, D = module.input0.shape
        samples = normal(0, 1, size=[mc_samples, N, D], device=module.input0.device)
        samples *= sqrt(2) / sqrt(mc_samples)

        if module.reduction == "mean":
            samples /= sqrt(module.input0.numel())

        return samples

    def _sum_hessian(self, module, g_inp, g_out):
        """The Hessian, summed across the batch dimension.

        Args:
            module: (torch.nn.MSELoss) module
            g_inp: Gradient of loss w.r.t. input
            g_out: Gradient of loss w.r.t. output

        Returns: a `[D, D]` tensor of the Hessian, summed across batch

        """
        self.check_input_dims(module)

        N, D = module.input0.shape
        H = 2 * eye(D, device=module.input0.device)

        if module.reduction == "sum":
            H *= N
        elif module.reduction == "mean":
            H /= D

        return H

    def _make_hessian_mat_prod(self, module, g_inp, g_out):
        """Multiplication of the input Hessian with a matrix."""

        def hessian_mat_prod(mat):
            Hmat = 2 * mat
            if module.reduction == "mean":
                Hmat /= module.input0.numel()
            return Hmat

        return hessian_mat_prod

    def check_input_dims(self, module):
        """Raises an exception if the shapes of the input are not supported."""
        if not len(module.input0.shape) == 2:
            raise ValueError("Only 2D inputs are currently supported for MSELoss.")

    def hessian_is_psd(self):
        return True



from torch import einsum


def extract_weight_diagonal(module, backproped):
    return einsum("vno,ni->oi", (backproped ** 2, module.input0 ** 2))


def extract_bias_diagonal(module, backproped):
    return einsum("vno->o", backproped ** 2)



from torch import eq
from torch.nn import Dropout

from backpack.core.derivatives.elementwise import ElementwiseDerivatives


class DropoutDerivatives(ElementwiseDerivatives):
    def get_module(self):
        return Dropout

    def hessian_is_zero(self):
        return True

    def df(self, module, g_inp, g_out):
        scaling = 1 / (1 - module.p)
        mask = 1 - eq(module.output, 0.0).float()
        return mask * scaling



""" Test implementation of `sqrt_hessian_sampled` for loss derivatives. """


import pytest
import torch
from torch.nn import CrossEntropyLoss, MSELoss

from backpack import extend

from .automated_test import check_sizes, check_values
from .test_sqrt_hessian import (
    autograd_input_hessian,
    classification_targets,
    derivative_from_layer,
    embed,
    generate_data_input_hessian,
    make_id,
    regression_targets,
    sample_hessians_from_sqrt,
)

torch.manual_seed(0)
ARGS = "layer,input_shape,targets,raises"
SETTINGS = [
    # reduction mean
    [
        # layer
        CrossEntropyLoss(reduction="mean"),
        # input_shape
        (2, 3),
        # targets
        classification_targets(shape=(2,), num_classes=3),
        # tuple of exceptions that should be raised
        (),
    ],
    [
        CrossEntropyLoss(reduction="mean"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (),
    ],
    # reduction sum
    [
        CrossEntropyLoss(reduction="sum"),
        (2, 3),
        classification_targets(shape=(2,), num_classes=3),
        (),
    ],
    [
        CrossEntropyLoss(reduction="sum"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (),
    ],
    # non-scalar outputs are not supported
    [
        CrossEntropyLoss(reduction="none"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (ValueError,),
    ],
    # no reduction for a single number as input should be fine
    [
        CrossEntropyLoss(reduction="none"),
        (1, 1),
        classification_targets(shape=(1,), num_classes=1),
        (),
    ],
    # reduction mean
    [MSELoss(reduction="mean"), (5, 1), regression_targets(shape=(5, 1)), ()],
    # reduction sum
    [MSELoss(reduction="sum"), (5, 1), regression_targets(shape=(5, 1)), ()],
    # no sampling implemented yet
    [
        MSELoss(reduction="sum"),
        (5, 2),
        regression_targets(shape=(5, 2)),
        (NotImplementedError,),
    ],
    [
        MSELoss(reduction="mean"),
        (5, 2),
        regression_targets(shape=(5, 2)),
        (NotImplementedError,),
    ],
    # non-scalar outputs are not supported
    [
        MSELoss(reduction="none"),
        (5, 1),
        regression_targets(shape=(5, 1)),
        (ValueError,),
    ],
]
IDS = [make_id(layer, input_shape) for (layer, input_shape, _, _) in SETTINGS]


def backpack_hessian_via_sqrt_hessian(layer, input, targets):
    layer = extend(layer)
    derivative = derivative_from_layer(layer)

    # forward pass to initialize backpack buffers
    _ = layer(input, targets)

    sqrt_hessian = derivative.sqrt_hessian(layer, None, None)
    individual_hessians = sample_hessians_from_sqrt(sqrt_hessian)

    hessian = embed(individual_hessians, input)

    return hessian


def backpack_hessian_via_sqrt_hessian_sampled(layer, input, targets):
    MC_SAMPLES = 100000

    layer = extend(layer)
    derivative = derivative_from_layer(layer)

    # forward pass to initialize backpack buffers
    _ = layer(input, targets)

    sqrt_hessian = derivative.sqrt_hessian_sampled(
        layer, None, None, mc_samples=MC_SAMPLES
    )
    individual_hessians = sample_hessians_from_sqrt(sqrt_hessian)

    # embed
    # TODO improve readability
    hessian_shape = (*input.shape, *input.shape)
    hessian = torch.zeros(hessian_shape)

    hessian = embed(individual_hessians, input)

    return hessian


@pytest.mark.parametrize(ARGS, SETTINGS, ids=IDS)
def test_sqrt_hessian_via_input_hessian_sampled(layer, input_shape, targets, raises):
    """Compare the Hessian to reconstruction from individual sampled Hessian sqrt."""
    torch.manual_seed(0)
    input = generate_data_input_hessian(input_shape)
    try:
        return _compare_input_hessian_sampled(layer, input, targets)
    except Exception as e:
        if isinstance(e, raises):
            return
        else:
            raise e


def _compare_input_hessian_sampled(layer, input, targets):
    RTOL, ATOL = 1e-2, 1e-2
    backpack_result = backpack_hessian_via_sqrt_hessian_sampled(layer, input, targets)
    autograd_result = autograd_input_hessian(layer, input, targets)

    check_sizes(autograd_result, backpack_result)
    check_values(autograd_result, backpack_result, rtol=RTOL, atol=ATOL)

    return backpack_result



"""The code relies on the insight that average pooling can be understood as
convolution over single channels with a constant kernel."""

import torch.nn
from backpack.core.derivatives.basederivatives import BaseDerivatives
from backpack.utils.ein import eingroup
from torch.nn import AvgPool2d, Conv2d, ConvTranspose2d


class AvgPool2DDerivatives(BaseDerivatives):
    def get_module(self):
        return AvgPool2d

    def hessian_is_zero(self):
        return True

    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        """Use fact that average pooling can be implemented as conv."""
        _, C, H_in, W_in = module.input0.size()
        in_features = C * H_in * W_in
        _, _, H_out, W_out = module.output.size()
        out_features = C * H_out * W_out

        mat = mat.reshape(out_features * C, 1, H_out, W_out)
        jac_t_mat = self.__apply_jacobian_t_of(module, mat).reshape(
            out_features, in_features
        )
        mat_t_jac = jac_t_mat.t().reshape(in_features * C, 1, H_out, W_out)
        jac_t_mat_t_jac = self.__apply_jacobian_t_of(module, mat_t_jac).reshape(
            in_features, in_features
        )

        return jac_t_mat_t_jac.t()

    def check_exotic_parameters(self, module):
        assert module.count_include_pad, (
            "Might not work for exotic hyperparameters of AvgPool2d, "
            + "like count_include_pad=False"
        )

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        self.check_exotic_parameters(module)

        mat_as_pool = self.__make_single_channel(mat, module)
        jmp_as_pool = self.__apply_jacobian_of(module, mat_as_pool)
        self.__check_jmp_out_as_pool(mat, jmp_as_pool, module)

        return self.reshape_like_output(jmp_as_pool, module)

    def __make_single_channel(self, mat, module):
        """Create fake single-channel images, grouping batch,
        class and channel dimension."""
        result = eingroup("v,n,c,w,h->vnc,w,h", mat)
        C_axis = 1
        return result.unsqueeze(C_axis)

    def __apply_jacobian_of(self, module, mat):
        conv2d = Conv2d(
            in_channels=1,
            out_channels=1,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            bias=False,
        ).to(module.input0.device)

        conv2d.weight.requires_grad = False
        avg_kernel = torch.ones_like(conv2d.weight) / conv2d.weight.numel()
        conv2d.weight.data = avg_kernel

        return conv2d(mat)

    def __check_jmp_out_as_pool(self, mat, jmp_as_pool, module):
        V = mat.size(0)
        N, C_out, H_out, W_out = module.output_shape
        assert jmp_as_pool.shape == (V * N * C_out, 1, H_out, W_out)

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        self.check_exotic_parameters(module)

        mat_as_pool = self.__make_single_channel(mat, module)
        jmp_as_pool = self.__apply_jacobian_t_of(module, mat_as_pool)
        self.__check_jmp_in_as_pool(mat, jmp_as_pool, module)

        return self.reshape_like_input(jmp_as_pool, module)

    def __apply_jacobian_t_of(self, module, mat):
        C_for_conv_t = 1

        conv2d_t = ConvTranspose2d(
            in_channels=C_for_conv_t,
            out_channels=C_for_conv_t,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            bias=False,
        ).to(module.input0.device)

        conv2d_t.weight.requires_grad = False
        avg_kernel = torch.ones_like(conv2d_t.weight) / conv2d_t.weight.numel()
        conv2d_t.weight.data = avg_kernel

        V_N_C_in = mat.size(0)
        _, _, H_in, W_in = module.input0.size()
        output_size = (V_N_C_in, C_for_conv_t, H_in, W_in)

        return conv2d_t(mat, output_size=output_size)

    def __check_jmp_in_as_pool(self, mat, jmp_as_pool, module):
        V = mat.size(0)
        N, C_in, H_in, W_in = module.input0_shape
        assert jmp_as_pool.shape == (V * N * C_in, 1, H_in, W_in)



class CTX:
    """
    Global Class holding the configuration of the backward pass
    """

    active_exts = tuple()
    debug = False

    @staticmethod
    def set_active_exts(active_exts):
        CTX.active_exts = tuple()
        for act_ext in active_exts:
            CTX.active_exts += (act_ext,)

    @staticmethod
    def get_active_exts():
        return CTX.active_exts

    @staticmethod
    def add_hook_handle(hook_handle):
        if getattr(CTX, "hook_handles", None) is None:
            CTX.hook_handles = []
        CTX.hook_handles.append(hook_handle)

    @staticmethod
    def remove_hooks():
        for handle in CTX.hook_handles:
            handle.remove()
        CTX.hook_handles = []

    @staticmethod
    def is_extension_active(extension_class):
        for backpack_ext in CTX.get_active_exts():
            if isinstance(backpack_ext, extension_class):
                return True
        return False

    @staticmethod
    def get_debug():
        return CTX.debug

    @staticmethod
    def set_debug(debug):
        CTX.debug = debug



from .hvp import hessian_vector_product
from .lop import L_op
from .rop import R_op


def ggn_vector_product(loss, output, model, v):
    """
    Multiplies the vector `v` with the Generalized Gauss-Newton,
    `ggn_v = J.T @ H @ J @ v`

    where `J` is the Jacobian of `output` w.r.t. `model.parameters()`
    and `H` is the Hessian of `loss` w.r.t. `output`.

    Example usage:
    ```
    X, Y = data()
    model = torch.nn.Linear(784, 10)
    lossfunc = torch.nn.CrossEntropyLoss()

    output = model(X)
    loss = lossfunc(output, Y)

    v = list([torch.randn_like(p) for p in model.parameters])

    GGNv = ggn_vector_product(loss, output, model, v)
    ```

    Parameters:
    -----------
        loss: torch.Tensor
        output: torch.Tensor
        model: torch.nn.Module
        v: [torch.Tensor]
            List of tensors matching the sizes of model.parameters()
    """
    return ggn_vector_product_from_plist(loss, output, list(model.parameters()), v)


def ggn_vector_product_from_plist(loss, output, plist, v):
    Jv = R_op(output, plist, v)
    HJv = hessian_vector_product(loss, output, Jv)
    JTHJv = L_op(output, plist, HJv)
    return JTHJv



from backpack.extensions.module_extension import ModuleExtension


class FirstOrderModuleExtension(ModuleExtension):
    def backpropagate(self, ext, module, g_inp, g_out, bpQuantities):
        return None



from torch import einsum
from torch.nn import Conv2d, ConvTranspose2d
from torch.nn.functional import conv2d

from backpack.core.derivatives.basederivatives import BaseParameterDerivatives
from backpack.utils import conv as convUtils
from backpack.utils.ein import eingroup


class Conv2DDerivatives(BaseParameterDerivatives):
    def get_module(self):
        return Conv2d

    def hessian_is_zero(self):
        return True

    def get_unfolded_input(self, module):
        return convUtils.unfold_func(module)(module.input0)

    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        _, C_in, H_in, W_in = module.input0.size()
        in_features = C_in * H_in * W_in
        _, C_out, H_out, W_out = module.output.size()
        out_features = C_out * H_out * W_out

        mat = mat.reshape(out_features, C_out, H_out, W_out)
        jac_t_mat = self.__jac_t(module, mat).reshape(out_features, in_features)

        mat_t_jac = jac_t_mat.t().reshape(in_features, C_out, H_out, W_out)
        jac_t_mat_t_jac = self.__jac_t(module, mat_t_jac).reshape(
            in_features, in_features
        )

        return jac_t_mat_t_jac.t()

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        mat_as_conv = eingroup("v,n,c,h,w->vn,c,h,w", mat)
        jmp_as_conv = conv2d(
            mat_as_conv,
            module.weight.data,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            groups=module.groups,
        )
        return self.reshape_like_output(jmp_as_conv, module)

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        mat_as_conv = eingroup("v,n,c,h,w->vn,c,h,w", mat)
        jmp_as_conv = self.__jac_t(module, mat_as_conv)
        return self.reshape_like_input(jmp_as_conv, module)

    def __jac_t(self, module, mat):
        """Apply Conv2d backward operation."""
        _, C_in, H_in, W_in = module.input0.size()
        _, C_out, H_out, W_out = module.output.size()
        H_axis = 2
        W_axis = 3

        conv2d_t = ConvTranspose2d(
            in_channels=C_out,
            out_channels=C_in,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            bias=False,
            dilation=module.dilation,
            groups=module.groups,
        ).to(module.input0.device)

        conv2d_t.weight.data = module.weight

        V_N = mat.size(0)
        output_size = (V_N, C_in, H_in, W_in)

        jac_t_mat = (
            conv2d_t(mat, output_size=output_size)
            .narrow(H_axis, 0, H_in)
            .narrow(W_axis, 0, W_in)
        )
        return jac_t_mat

    def _bias_jac_mat_prod(self, module, g_inp, g_out, mat):
        """mat has shape [V, C_out]"""
        # expand for each batch and for each channel
        N_axis, H_axis, W_axis = 1, 3, 4
        jac_mat = mat.unsqueeze(N_axis).unsqueeze(H_axis).unsqueeze(W_axis)

        N, _, H_out, W_out = module.output_shape
        return jac_mat.expand(-1, N, -1, H_out, W_out)

    def _bias_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        N_axis, H_axis, W_axis = 1, 3, 4
        axes = [H_axis, W_axis]
        if sum_batch:
            axes = [N_axis] + axes

        return mat.sum(axes)

    # TODO: Improve performance by using conv instead of unfold

    def _weight_jac_mat_prod(self, module, g_inp, g_out, mat):
        jac_mat = eingroup("v,o,i,h,w->v,o,ihw", mat)
        X = self.get_unfolded_input(module)

        jac_mat = einsum("nij,vki->vnkj", (X, jac_mat))
        return self.reshape_like_output(jac_mat, module)

    def _weight_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        """Unintuitive, but faster due to convolution."""
        V = mat.shape[0]
        N, C_out, _, _ = module.output_shape
        _, C_in, _, _ = module.input0_shape

        mat = eingroup("v,n,c,w,h->vn,c,w,h", mat).repeat(1, C_in, 1, 1)
        C_in_axis = 1
        # a,b represent the combined/repeated dimensions
        mat = eingroup("a,b,w,h->ab,w,h", mat).unsqueeze(C_in_axis)

        N_axis = 0
        input = eingroup("n,c,h,w->nc,h,w", module.input0).unsqueeze(N_axis)
        input = input.repeat(1, V, 1, 1)

        grad_weight = conv2d(
            input,
            mat,
            bias=None,
            stride=module.dilation,
            padding=module.padding,
            dilation=module.stride,
            groups=C_in * N * V,
        ).squeeze(0)

        K_H_axis, K_W_axis = 1, 2
        _, _, K_H, K_W = module.weight.shape
        grad_weight = grad_weight.narrow(K_H_axis, 0, K_H).narrow(K_W_axis, 0, K_W)

        eingroup_eq = "vnio,x,y->v,{}o,i,x,y".format("" if sum_batch else "n,")
        return eingroup(
            eingroup_eq, grad_weight, dim={"v": V, "n": N, "i": C_in, "o": C_out}
        )



from backpack.extensions.mat_to_mat_jac_base import MatToJacMat


class DiagGGNBaseModule(MatToJacMat):
    def __init__(self, derivatives, params=None):
        super().__init__(derivatives, params=params)



from .module_extension import ModuleExtension


class MatToJacMat(ModuleExtension):
    """
    Base class for backpropagating matrices by multiplying with Jacobians.
    """

    def __init__(self, derivatives, params=None):
        super().__init__(params)
        self.derivatives = derivatives

    def backpropagate(self, ext, module, grad_inp, grad_out, backproped):

        if isinstance(backproped, list):
            M_list = [
                self.derivatives.jac_t_mat_prod(module, grad_inp, grad_out, M)
                for M in backproped
            ]
            return list(M_list)
        else:
            return self.derivatives.jac_t_mat_prod(
                module, grad_inp, grad_out, backproped
            )



"""An example to check if BackPACK' first-order extensions are working for ResNets."""

import torch

from backpack import backpack, extend, extensions

from .automated_test import check_sizes, check_values
from .test_sum_hessian import classification_targets


def autograd_individual_gradients(X, y, model, loss_func):
    """Individual gradients via for loop with automatic differentiation.

    Args:
        X (torch.Tensor): Mini-batch of shape `(N, *)`
        y (torch.Tensor: Labels for `X`
        model (torch.nn.Module): Model for forward pass
        loss_func (torch.nn.Module): Loss function for model prediction

    Returns:
        [torch.Tensor]: Individual gradients for samples in the mini-batch
            with respect to the model parameters. Arranged in the same order
           as `model.parameters()`.
    """
    N = X.shape[0]
    reduction_factor = _get_reduction_factor(X, y, model, loss_func)

    individual_gradients = [
        torch.zeros(N, *p.shape).to(X.device) for p in model.parameters()
    ]

    for n in range(N):
        x_n = X[n].unsqueeze(0)
        y_n = y[n].unsqueeze(0)

        f_n = model(x_n)
        l_n = loss_func(f_n, y_n) / reduction_factor

        g_n = torch.autograd.grad(l_n, model.parameters())

        for idx, g in enumerate(g_n):
            individual_gradients[idx][n] = g

    return individual_gradients


def _get_reduction_factor(X, y, model, loss_func):
    """Return reduction factor of loss function."""
    N = X.shape[0]

    x_0 = X[0].unsqueeze(0)
    y_0 = y[0].unsqueeze(0)

    x_0_repeated = x_0.repeat([N if pos == 0 else 1 for pos, _ in enumerate(X.shape)])
    y_0_repeated = y_0.repeat([N if pos == 0 else 1 for pos, _ in enumerate(y.shape)])

    individual_loss = loss_func(model(x_0), y_0)
    reduced_loss = loss_func(model(x_0_repeated), y_0_repeated)

    return (N * individual_loss / reduced_loss).item()


def backpack_individual_gradients(X, y, model, loss_func):
    """Individual gradients with BackPACK.

    Args:
        X (torch.Tensor): Mini-batch of shape `(N, *)`
        y (torch.Tensor: Labels for `X`
        model (torch.nn.Module): Model for forward pass
        loss_func (torch.nn.Module): Loss function for model prediction

    Returns:
        [torch.Tensor]: Individual gradients for samples in the mini-batch
            with respect to the model parameters. Arranged in the same order
           as `model.parameters()`.
    """
    model = extend(model)
    loss_func = extend(loss_func)

    loss = loss_func(model(X), y)

    with backpack(extensions.BatchGrad()):
        loss.backward()

    individual_gradients = [p.grad_batch for p in model.parameters()]

    return individual_gradients


class Identity(torch.nn.Module):
    """Identity operation."""

    def forward(self, input):
        return input


class Parallel(torch.nn.Sequential):
    """Feed input to multiple modules, sum the result.

              |-----|
        | ->  | f_1 |  -> |
        |     |-----|     |
        |                 |
        |     |-----|     |
    x ->| ->  | f_2 |  -> + -> f₁(x) + f₂(x) + ...
        |     |-----|     |
        |                 |
        |     |-----|     |
        | ->  | ... |  -> |
              |-----|

    """

    def forward(self, input):
        """Process input with all modules, sum the output."""
        for idx, module in enumerate(self.children()):
            if idx == 0:
                output = module(input)
            else:
                output = output + module(input)

        return output


def test_individual_gradients_simple_resnet():
    """Individual gradients for a simple ResNet with autodiff and BackPACK."""

    # batch size, feature dimension
    N, D = 2, 5
    # classification
    C = 3

    X = torch.rand(N, D)
    y = classification_targets((N,), num_classes=C)

    model = Parallel(Identity(), torch.nn.Linear(D, D, bias=True))
    loss_func = torch.nn.CrossEntropyLoss(reduction="sum")

    result_autograd = autograd_individual_gradients(X, y, model, loss_func)
    result_backpack = backpack_individual_gradients(X, y, model, loss_func)

    check_sizes(result_autograd, result_backpack)
    check_values(result_autograd, result_backpack)



from torch import einsum

from backpack.core.derivatives.shape_check import (
    R_mat_prod_accept_vectors,
    R_mat_prod_check_shapes,
)
from backpack.extensions.curvature import Curvature
from backpack.extensions.module_extension import ModuleExtension


class CMPBase(ModuleExtension):
    """
    Given matrix-vector product routine `MVP(A)` backpropagate to `MVP(J^T A J)`.
    """

    def __init__(self, derivatives, params=None):
        super().__init__(params=params)
        self.derivatives = derivatives

    def backpropagate(self, ext, module, g_inp, g_out, backproped):
        """Backpropagate Hessian multiplication routines.

        Given mat → ℋz(x) mat, backpropagate mat → ℋx mat.
        """
        GGN_mat_prod = self._make_GGN_mat_prod(ext, module, g_inp, g_out, backproped)

        R_required = self._require_residual(ext, module, g_inp, g_out, backproped)
        if R_required:
            R_mat_prod = self._make_R_mat_prod(ext, module, g_inp, g_out, backproped)

        def CMP_in(mat):
            """Multiplication with curvature matrix w.r.t. the module input.

            Parameters:
            -----------
            mat : torch.Tensor
                Matrix that will be multiplied.
            """
            out = GGN_mat_prod(mat)

            if R_required:
                out.add_(R_mat_prod(mat))

            return out

        return CMP_in

    def _make_GGN_mat_prod(self, ext, module, g_inp, g_out, backproped):
        """Return multiplication routine with the first HBP term."""
        CMP_out = backproped

        def GGN_mat_prod(mat):
            """Multiply with the GGN term: mat → [𝒟z(x)ᵀ ℋz 𝒟z(x)] mat.

            First term of the module input Hessian backpropagation equation.
            """
            Jmat = self.derivatives.jac_mat_prod(module, g_inp, g_out, mat)
            CJmat = CMP_out(Jmat)
            JTCJmat = self.derivatives.jac_t_mat_prod(module, g_inp, g_out, CJmat)

            return JTCJmat

        return GGN_mat_prod

    def _require_residual(self, ext, module, g_inp, g_out, backproped):
        """Is the residual term required for multiply with the curvature?"""
        vanishes = self.derivatives.hessian_is_zero()
        neglected = not Curvature.require_residual(ext.get_curv_type())

        return not (vanishes or neglected)

    def _make_R_mat_prod(self, ext, module, g_inp, g_out, backproped):
        """Return multiplication routine with the second HBP term."""
        if self.derivatives.hessian_is_diagonal():
            R_mat_prod = self.__make_diagonal_R_mat_prod(
                ext, module, g_inp, g_out, backproped
            )
        else:
            R_mat_prod = self.__make_nondiagonal_R_mat_prod(
                ext, module, g_inp, g_out, backproped
            )

        return R_mat_prod

    def __make_diagonal_R_mat_prod(self, ext, module, g_inp, g_out, backproped):
        # TODO Refactor core: hessian_diagonal -> residual_diagonal
        R = self.derivatives.hessian_diagonal(module, g_inp, g_out)
        R_mod = Curvature.modify_residual(R, ext.get_curv_type())

        @R_mat_prod_accept_vectors
        @R_mat_prod_check_shapes
        def make_residual_mat_prod(self, module, g_inp, g_out):
            def R_mat_prod(mat):
                """Multiply with the residual: mat → [∑_{k} Hz_k(x) 𝛿z_k] mat.

                Second term of the module input Hessian backpropagation equation.
                """
                return einsum("n...,vn...->vn...", (R_mod, mat))

            return R_mat_prod

        return make_residual_mat_prod(self, module, g_inp, g_out)

    def __make_nondiagonal_R_mat_prod(self, ext, module, g_inp, g_out, backproped):
        curv_type = ext.get_curv_type()
        if not Curvature.is_pch(curv_type):
            R_mat_prod = self.derivatives.make_residual_mat_prod(module, g_inp, g_out)
        else:
            raise ValueError(
                "{} not supported for {}. Residual cannot be cast PSD.".format(
                    curv_type, module
                )
            )

        return R_mat_prod



import torch
import torch.nn

from backpack.core.derivatives.conv2d import Conv2DDerivatives
from backpack.extensions.secondorder.diag_hessian.diag_h_base import DiagHBaseModule
from backpack.utils import conv as convUtils


class DiagHConv2d(DiagHBaseModule):
    def __init__(self):
        super().__init__(derivatives=Conv2DDerivatives(), params=["bias", "weight"])

    def bias(self, ext, module, g_inp, g_out, backproped):
        sqrt_h_outs = backproped["matrices"]
        sqrt_h_outs_signs = backproped["signs"]
        h_diag = torch.zeros_like(module.bias)

        for h_sqrt, sign in zip(sqrt_h_outs, sqrt_h_outs_signs):
            h_diag_curr = convUtils.extract_bias_diagonal(module, h_sqrt)
            h_diag.add_(sign * h_diag_curr)
        return h_diag

    def weight(self, ext, module, g_inp, g_out, backproped):
        sqrt_h_outs = backproped["matrices"]
        sqrt_h_outs_signs = backproped["signs"]
        X = convUtils.unfold_func(module)(module.input0)
        h_diag = torch.zeros_like(module.weight)

        for h_sqrt, sign in zip(sqrt_h_outs, sqrt_h_outs_signs):
            h_diag_curr = convUtils.extract_weight_diagonal(module, X, h_sqrt)
            h_diag.add_(sign * h_diag_curr)
        return h_diag



from torch import einsum

from backpack.core.derivatives.basederivatives import BaseDerivatives


class ElementwiseDerivatives(BaseDerivatives):
    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        self._no_inplace(module)

        df_elementwise = self.df(module, g_inp, g_out)
        return einsum("...,v...->v...", (df_elementwise, mat))

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        self._no_inplace(module)

        return self.jac_t_mat_prod(module, g_inp, g_out, mat)

    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        self._no_inplace(module)

        batch, df_flat = self.batch_flat(self.df(module, g_inp, g_out))
        return einsum("ni,nj,ij->ij", (df_flat, df_flat, mat)) / batch

    def hessian_diagonal(self, module, g_inp, g_out):
        self._no_inplace(module)

        return self.d2f(module, g_inp, g_out) * g_out[0]

    def df(self, module, g_inp, g_out):
        raise NotImplementedError("First derivatives not implemented")

    def d2f(self, module, g_inp, g_out):
        raise NotImplementedError("Second derivatives not implemented")

    @staticmethod
    def _no_inplace(module):
        """Do not support inplace modification.

        Jacobians/Hessians might be computed using the modified input instead
        of the original.

        Args:
            module (torch.nn.Module): Elementwise activation module.

        Raises:
            NotImplementedError: If `module` has inplace option enabled.

        Todo:
            - Write tests to investigate what happens with `inplace=True`.
        """
        has_inplace_option = hasattr(module, "inplace")

        if has_inplace_option:
            if module.inplace is True:
                raise NotImplementedError("Inplace not supported in {}.".format(module))



"""
Curvature-matrix product backPACK extensions.

Those extension propagate additional information through the computation graph.
They are more expensive to run than a standard gradient backpropagation.

This extension does not compute information directly, but gives access to
functions to compute Matrix-Matrix products with Block-Diagonal approximations
of the curvature, such as the Block-diagonal Generalized Gauss-Newton
"""

from torch.nn import (
    AvgPool2d,
    BatchNorm1d,
    Conv2d,
    CrossEntropyLoss,
    Dropout,
    Flatten,
    Linear,
    MaxPool2d,
    MSELoss,
    ReLU,
    Sigmoid,
    Tanh,
    ZeroPad2d,
)

from backpack.extensions.backprop_extension import BackpropExtension

from . import (
    activations,
    batchnorm1d,
    conv2d,
    dropout,
    flatten,
    linear,
    losses,
    padding,
    pooling,
)


class CMP(BackpropExtension):
    def __init__(self, curv_type, savefield="cmp"):
        self.curv_type = curv_type

        super().__init__(
            savefield=savefield,
            fail_mode="ERROR",
            module_exts={
                MSELoss: losses.CMPMSELoss(),
                CrossEntropyLoss: losses.CMPCrossEntropyLoss(),
                Linear: linear.CMPLinear(),
                MaxPool2d: pooling.CMPMaxpool2d(),
                AvgPool2d: pooling.CMPAvgPool2d(),
                ZeroPad2d: padding.CMPZeroPad2d(),
                Conv2d: conv2d.CMPConv2d(),
                Dropout: dropout.CMPDropout(),
                Flatten: flatten.CMPFlatten(),
                ReLU: activations.CMPReLU(),
                Sigmoid: activations.CMPSigmoid(),
                Tanh: activations.CMPTanh(),
                BatchNorm1d: batchnorm1d.CMPBatchNorm1d(),
            },
        )

    def get_curv_type(self):
        return self.curv_type



from backpack.core.derivatives.crossentropyloss import CrossEntropyLossDerivatives
from backpack.core.derivatives.mseloss import MSELossDerivatives
from backpack.extensions.curvature import Curvature
from backpack.extensions.curvmatprod.cmpbase import CMPBase


class CMPLoss(CMPBase):
    def backpropagate(self, ext, module, g_inp, g_out, backproped):
        Curvature.check_loss_hessian(
            self.derivatives.hessian_is_psd(), curv_type=ext.get_curv_type()
        )

        CMP = self.derivatives.make_hessian_mat_prod(module, g_inp, g_out)
        return CMP


class CMPMSELoss(CMPLoss):
    def __init__(self):
        super().__init__(derivatives=MSELossDerivatives())


class CMPCrossEntropyLoss(CMPLoss):
    def __init__(self):
        super().__init__(derivatives=CrossEntropyLossDerivatives())



from backpack.core.derivatives.dropout import DropoutDerivatives

from .cmpbase import CMPBase


class CMPDropout(CMPBase):
    def __init__(self):
        super().__init__(derivatives=DropoutDerivatives())



"""Test implementation of `sqrt_hessian` for loss derivatives."""


import pytest
import torch
from torch.nn import CrossEntropyLoss, MSELoss

from backpack import extend
from backpack.core.derivatives import derivatives_for
from backpack.hessianfree.hvp import hessian_vector_product

from .automated_test import check_sizes, check_values


def classification_targets(shape, num_classes):
    """Create random targets for classes 0, ..., `num_classes - 1`."""
    return torch.randint(size=shape, low=0, high=num_classes)


def regression_targets(shape):
    """Create random targets for regression."""
    return torch.rand(size=shape)


def make_id(layer, input_shape):
    """Create an id for the test scenario."""
    return "in{}-{}".format(input_shape, layer)


torch.manual_seed(0)
ARGS = "layer,input_shape,targets,raises"
SETTINGS = [
    # reduction mean
    [
        # layer
        CrossEntropyLoss(reduction="mean"),
        # input_shape
        (2, 3),
        # targets
        classification_targets(shape=(2,), num_classes=3),
        # tuple of exceptions that should be raised
        (),
    ],
    [
        CrossEntropyLoss(reduction="mean"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (),
    ],
    # reduction sum
    [
        CrossEntropyLoss(reduction="sum"),
        (2, 3),
        classification_targets(shape=(2,), num_classes=3),
        (),
    ],
    [
        CrossEntropyLoss(reduction="sum"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (),
    ],
    # non-scalar outputs are not supported
    [
        CrossEntropyLoss(reduction="none"),
        (8, 2),
        classification_targets(shape=(8,), num_classes=2),
        (ValueError,),
    ],
    # no reduction for a single number as input should be fine
    [
        CrossEntropyLoss(reduction="none"),
        (1, 1),
        classification_targets(shape=(1,), num_classes=1),
        (),
    ],
    # reduction mean
    [MSELoss(reduction="mean"), (5, 1), regression_targets(shape=(5, 1)), ()],
    [MSELoss(reduction="mean"), (5, 2), regression_targets(shape=(5, 2)), (),],
    # reduction sum
    [MSELoss(reduction="sum"), (5, 1), regression_targets(shape=(5, 1)), ()],
    [MSELoss(reduction="sum"), (5, 2), regression_targets(shape=(5, 2)), (),],
    # non-scalar outputs are not supported
    [
        MSELoss(reduction="none"),
        (5, 1),
        regression_targets(shape=(5, 1)),
        (ValueError,),
    ],
]
IDS = [make_id(layer, input_shape) for (layer, input_shape, _, _) in SETTINGS]


def autograd_hessian(loss, x):
    """Return the Hessian matrix of `loss` w.r.t. `x`.

    Arguments:
        loss (torch.Tensor): A scalar-valued tensor.
        x (torch.Tensor): Tensor used in the computation graph of `loss`.
    Shapes:
        loss: `[1,]`
        x: `[A, B, C, ...]`
    Returns:
        torch.Tensor: Hessian tensor of `loss` w.r.t. `x`. The Hessian has shape
            `[A, B, C, ..., A, B, C, ...]`.
    """
    assert loss.numel() == 1

    vectorized_shape = (x.numel(), x.numel())
    final_shape = (*x.shape, *x.shape)

    hessian_vec_x = torch.zeros(vectorized_shape)

    num_cols = hessian_vec_x.shape[1]
    for column_idx in range(num_cols):
        unit = torch.zeros(num_cols)
        unit[column_idx] = 1.0

        unit = unit.view_as(x)
        column = hessian_vector_product(loss, [x], [unit])[0].reshape(-1)

        hessian_vec_x[:, column_idx] = column

    return hessian_vec_x.reshape(final_shape)


def autograd_input_hessian(layer, input, targets):
    """Compute the Hessian of a loss module w.r.t. its input."""
    input.requires_grad = True

    loss = layer(input, targets)
    return autograd_hessian(loss, input)


# TODO remove duplicate
def derivative_from_layer(layer):
    layer_to_derivative = derivatives_for

    for module_cls, derivative_cls in layer_to_derivative.items():
        if isinstance(layer, module_cls):
            return derivative_cls()

    raise RuntimeError("No derivative available for {}".format(layer))


def embed(individual_hessians, input):
    hessian_shape = (*input.shape, *input.shape)
    hessian = torch.zeros(hessian_shape)

    N = input.shape[0]

    for n in range(N):
        num_axes = len(input.shape)

        if num_axes == 2:
            hessian[n, :, n, :] = individual_hessians[n]
        else:
            raise ValueError("Only 2D inputs are currently supported.")

    return hessian


def backpack_hessian_via_sqrt_hessian(layer, input, targets):
    layer = extend(layer)
    derivative = derivative_from_layer(layer)

    # forward pass to initialize backpack buffers
    _ = layer(input, targets)

    sqrt_hessian = derivative.sqrt_hessian(layer, None, None)
    individual_hessians = sample_hessians_from_sqrt(sqrt_hessian)

    hessian = embed(individual_hessians, input)

    return hessian


def sample_hessians_from_sqrt(sqrt):
    """Convert individual matrix square root into individual full matrix. """
    equation = None
    num_axes = len(sqrt.shape)

    # TODO improve readability
    if num_axes == 3:
        equation = "vni,vnj->nij"
    else:
        raise ValueError("Only 2D inputs are currently supported.")

    return torch.einsum(equation, sqrt, sqrt)


def generate_data_input_hessian(input_shape):
    input = torch.rand(input_shape)
    return input


@pytest.mark.parametrize(ARGS, SETTINGS, ids=IDS)
def test_sqrt_hessian_via_input_hessian(layer, input_shape, targets, raises):
    """Compare the Hessian to reconstruction from individual Hessian sqrt."""
    torch.manual_seed(0)
    input = generate_data_input_hessian(input_shape)
    try:
        _compare_hessian_via_sqrt_hessian(layer, input, targets)
    except Exception as e:
        if not isinstance(e, raises):
            raise e


def _compare_hessian_via_sqrt_hessian(layer, input, targets):
    backpack_result = backpack_hessian_via_sqrt_hessian(layer, input, targets)
    autograd_result = autograd_input_hessian(layer, input, targets)

    check_sizes(autograd_result, backpack_result)
    check_values(autograd_result, backpack_result)

    return backpack_result



import warnings

from torch.nn import Sequential

FAIL_ERROR = "ERROR"
FAIL_WARN = "WARN"
FAIL_SILENT = "SILENT"


class BackpropExtension:
    """
    Base class for the BackPACK extensions.

    Descendants of this class need to
    - define in what field to save results
    - provide a mapping from Module classes to ModuleExtension instances.

    They can then be passed to the Backpack context manager, i.e.,
    ```
    with backpack(NewPackpropExtension("myfield", module_to_extensions)):
        loss(model(X), Y).backward()

        for p in model.parameters():
            print(p.myfield)
    ```
    """

    __external_module_extensions = {}

    def __init__(self, savefield, module_exts, fail_mode=FAIL_ERROR):
        """
        Parameters
        ----------
        savefield : str
            Where to save results
        module_exts : dict
            Dictionary mapping module classes to `ModuleExtension` instances
        fail_mode : str, optional
            Behavior when encountering an unknown layer.
            Can be
            - "ERROR": raise a NotImplementedError
            - "WARN": raise a UserWarning
            - "SILENT": skip the module silently
        """
        self.savefield = savefield
        self.__module_extensions = {
            **module_exts,
            **self.__class__.__external_module_extensions,
        }

        self.__fail_mode = fail_mode

    @classmethod
    def add_module_extension(cls, module, extension):
        cls.__external_module_extensions[module] = extension

    def __get_module_extension(self, module):
        module_extension = self.__module_extensions.get(module.__class__)

        def no_op(*args):
            return None

        if module_extension is None:

            if isinstance(module, Sequential):
                return no_op

            if self.__fail_mode is FAIL_ERROR:
                raise NotImplementedError(
                    "Extension saving to {} ".format(self.savefield)
                    + "does not have an extension for "
                    + "Module {}".format(module.__class__)
                )
            elif self.__fail_mode == FAIL_WARN:
                warnings.warn(
                    "Extension saving to {} ".format(self.savefield)
                    + "does not have an extension for "
                    + "Module {}".format(module.__class__)
                )

            return no_op

        return module_extension.apply

    def apply(self, module, g_inp, g_out):
        module_extension = self.__get_module_extension(module)
        module_extension(self, module, g_inp, g_out)



from backpack.core.derivatives.batchnorm1d import BatchNorm1dDerivatives
from backpack.extensions.firstorder.batch_grad.batch_grad_base import BatchGradBase


class BatchGradBatchNorm1d(BatchGradBase):
    def __init__(self):
        super().__init__(
            derivatives=BatchNorm1dDerivatives(), params=["bias", "weight"]
        )



"""Test KFRA backpropagation

H_in →  1/N ∑ₙ Jₙ^T H_out Jₙ

Notes:
    - `Dropout` cannot be tested,as the `autograd` implementation does a forward
       pass over each sample, while the `backpack` implementation requires only
       one forward pass over the batched data. This leads to different outputs,
       as `Dropout` is not deterministic.
"""
import pytest
import torch
from torch.nn import AvgPool2d, Conv2d, Linear, MaxPool2d, Sigmoid, Tanh, ZeroPad2d

from backpack import extend
from backpack.core.derivatives import derivatives_for
from backpack.hessianfree.lop import transposed_jacobian_vector_product

from .automated_test import check_sizes, check_values


def make_id(layer, input_shape):
    return "in{}-{}".format(input_shape, layer)


torch.manual_seed(0)
ARGS = "layer,input_shape"
SETTINGS = [
    # (layer, input_shape)
    [Linear(20, 10), (5, 20)],
    [MaxPool2d(kernel_size=2), (5, 3, 10, 8)],
    [MaxPool2d(kernel_size=2), (1, 2, 4, 4)],
    [MaxPool2d(kernel_size=3, stride=2, padding=1), (3, 2, 9, 11)],
    [AvgPool2d(kernel_size=3), (3, 2, 15, 13)],
    [AvgPool2d(kernel_size=3, stride=2), (3, 2, 15, 13)],
    [AvgPool2d(kernel_size=3, stride=2, padding=1), (3, 2, 15, 13)],
    [Sigmoid(), (6, 20)],
    [Sigmoid(), (6, 2, 7)],
    [Tanh(), (6, 20)],
    [Tanh(), (6, 2, 7)],
    [Conv2d(2, 3, kernel_size=2), (3, 2, 11, 13)],
    [Conv2d(2, 3, kernel_size=2, padding=1), (3, 2, 11, 13)],
    [Conv2d(2, 3, kernel_size=2, padding=1, stride=2), (3, 2, 11, 13)],
    [Conv2d(2, 3, kernel_size=2, padding=1, stride=2, dilation=2), (3, 2, 11, 13)],
    [ZeroPad2d(2), (4, 3, 4, 5)],
]
IDS = [make_id(layer, input_shape) for (layer, input_shape) in SETTINGS]


def get_output_shape(input, layer):
    return layer(input).shape


def autograd_ea_jac_t_mat_jac_prod(layer, input, mat):
    def sample_jac_t_mat_jac_prod(layer, sample, mat):
        assert sample.shape[0] == 1, "input is not batch size 1: {}".format(
            sample.shape
        )
        assert len(mat.shape) == 2

        def sample_jac_t_mat_prod(layer, sample, mat):
            result = torch.zeros(sample.numel(), mat.size(1))

            sample.requires_grad = True
            output = layer(sample)

            for col in range(mat.size(1)):
                column = mat[:, col].reshape(output.shape)
                result[:, col] = transposed_jacobian_vector_product(
                    [output], [sample], [column], retain_graph=True
                )[0].reshape(-1)

            return result

        jac_t_mat = sample_jac_t_mat_prod(layer, sample, mat)
        mat_t_jac = jac_t_mat.t()
        jac_t_mat_t_jac = sample_jac_t_mat_prod(layer, sample, mat_t_jac)
        jac_t_mat_jac = jac_t_mat_t_jac.t()

        return jac_t_mat_jac

    N = input.shape[0]
    input_features = input.shape.numel() // N

    result = torch.zeros(input_features, input_features)

    for n in range(N):
        sample_n = input[n].unsqueeze(0)
        result += sample_jac_t_mat_jac_prod(layer, sample_n, mat)

    return result / N


def derivative_from_layer(layer):
    layer_to_derivative = derivatives_for

    for module_cls, derivative_cls in layer_to_derivative.items():
        if isinstance(layer, module_cls):
            return derivative_cls()

    raise RuntimeError("No derivative available for {}".format(layer))


def backpack_ea_jac_t_mat_jac_prod(layer, input, mat):
    layer = extend(layer)
    derivative = derivative_from_layer(layer)

    # forward pass to initialize backpack buffers
    _ = layer(input)

    return derivative.ea_jac_t_mat_jac_prod(layer, None, None, mat)


def generate_data_ea_jac_t_mat_jac_prod(layer, input_shape):
    input = torch.rand(input_shape)
    out_features = get_output_shape(input, layer)[1:].numel()
    mat = torch.rand(out_features, out_features)
    return input, mat


@pytest.mark.parametrize(ARGS, SETTINGS, ids=IDS)
def test_ea_jac_t_mat_jac_prod(layer, input_shape):
    torch.manual_seed(0)
    input, mat = generate_data_ea_jac_t_mat_jac_prod(layer, input_shape)
    return _compare_ea_jac_t_mat_jac_prod(layer, input, mat)


def _compare_ea_jac_t_mat_jac_prod(layer, input, mat):
    autograd_result = autograd_ea_jac_t_mat_jac_prod(layer, input, mat)
    backpack_result = backpack_ea_jac_t_mat_jac_prod(layer, input, mat)

    check_sizes(autograd_result, backpack_result)
    check_values(autograd_result, backpack_result)

    return backpack_result


def test_ea_jac_t_mat_jac_prod_linear_manual():
    # Linear with manual ea_jac_t_mat_jac_prod
    input_shape = (5, 13)
    layer = torch.nn.Linear(13, 10)

    input, mat = generate_data_ea_jac_t_mat_jac_prod(layer, input_shape)

    test_result = _compare_ea_jac_t_mat_jac_prod(layer, input, mat)
    manual_result = torch.einsum(
        "ab,ac,cd->bd", layer.weight.data, mat, layer.weight.data
    )

    check_sizes(test_result, manual_result)
    check_values(test_result, manual_result)



from torch.nn import (
    AvgPool2d,
    Conv2d,
    CrossEntropyLoss,
    Dropout,
    Flatten,
    Linear,
    MaxPool2d,
    MSELoss,
    ReLU,
    Sigmoid,
    Tanh,
    ZeroPad2d,
)

from backpack.extensions.backprop_extension import BackpropExtension

from . import activations, conv2d, dropout, flatten, linear, losses, padding, pooling


class DiagHessian(BackpropExtension):
    """
    Diagonal of the Hessian.

    Stores the output in :code:`diag_h`, has the same dimensions as the gradient.

    .. warning::

        Very expensive on networks with non-piecewise linear activations.

    """

    def __init__(self):
        super().__init__(
            savefield="diag_h",
            fail_mode="ERROR",
            module_exts={
                MSELoss: losses.DiagHMSELoss(),
                CrossEntropyLoss: losses.DiagHCrossEntropyLoss(),
                Linear: linear.DiagHLinear(),
                MaxPool2d: pooling.DiagHMaxPool2d(),
                AvgPool2d: pooling.DiagHAvgPool2d(),
                ZeroPad2d: padding.DiagHZeroPad2d(),
                Conv2d: conv2d.DiagHConv2d(),
                Dropout: dropout.DiagHDropout(),
                Flatten: flatten.DiagHFlatten(),
                ReLU: activations.DiagHReLU(),
                Sigmoid: activations.DiagHSigmoid(),
                Tanh: activations.DiagHTanh(),
            },
        )



from torch.nn import (
    AvgPool2d,
    Conv2d,
    CrossEntropyLoss,
    Dropout,
    Flatten,
    Linear,
    MaxPool2d,
    MSELoss,
    ReLU,
    Sigmoid,
    Tanh,
    ZeroPad2d,
)

from backpack.extensions.backprop_extension import BackpropExtension
from backpack.extensions.secondorder.hbp import LossHessianStrategy

from . import activations, conv2d, dropout, flatten, linear, losses, padding, pooling


class DiagGGN(BackpropExtension):
    VALID_LOSS_HESSIAN_STRATEGIES = [
        LossHessianStrategy.EXACT,
        LossHessianStrategy.SAMPLING,
    ]

    def __init__(self, loss_hessian_strategy=LossHessianStrategy.EXACT, savefield=None):
        if savefield is None:
            savefield = "diag_ggn"
        if loss_hessian_strategy not in self.VALID_LOSS_HESSIAN_STRATEGIES:
            raise ValueError(
                "Unknown hessian strategy: {}".format(loss_hessian_strategy)
                + "Valid strategies: [{}]".format(self.VALID_LOSS_HESSIAN_STRATEGIES)
            )

        self.loss_hessian_strategy = loss_hessian_strategy
        super().__init__(
            savefield=savefield,
            fail_mode="ERROR",
            module_exts={
                MSELoss: losses.DiagGGNMSELoss(),
                CrossEntropyLoss: losses.DiagGGNCrossEntropyLoss(),
                Linear: linear.DiagGGNLinear(),
                MaxPool2d: pooling.DiagGGNMaxPool2d(),
                AvgPool2d: pooling.DiagGGNAvgPool2d(),
                ZeroPad2d: padding.DiagGGNZeroPad2d(),
                Conv2d: conv2d.DiagGGNConv2d(),
                Dropout: dropout.DiagGGNDropout(),
                Flatten: flatten.DiagGGNFlatten(),
                ReLU: activations.DiagGGNReLU(),
                Sigmoid: activations.DiagGGNSigmoid(),
                Tanh: activations.DiagGGNTanh(),
            },
        )


class DiagGGNExact(DiagGGN):
    """
    Diagonal of the Generalized Gauss-Newton/Fisher. 
    Uses the exact Hessian of the loss w.r.t. the model output.

    Stores the output in :code:`diag_ggn_exact`, 
    has the same dimensions as the gradient.

    For a faster but less precise alternative, 
    see :py:meth:`backpack.extensions.DiagGGNMC`.

    """

    def __init__(self):
        super().__init__(
            loss_hessian_strategy=LossHessianStrategy.EXACT, savefield="diag_ggn_exact"
        )


class DiagGGNMC(DiagGGN):
    """
    Diagonal of the Generalized Gauss-Newton/Fisher.
    Uses a Monte-Carlo approximation of
    the Hessian of the loss w.r.t. the model output.

    Stores the output in :code:`diag_ggn_mc`,
    has the same dimensions as the gradient.

    For a more precise but slower alternative,
    see :py:meth:`backpack.extensions.DiagGGNExact`.

    """

    def __init__(self, mc_samples=1):
        self._mc_samples = mc_samples
        super().__init__(
            loss_hessian_strategy=LossHessianStrategy.SAMPLING, savefield="diag_ggn_mc"
        )

    def get_num_mc_samples(self):
        return self._mc_samples



from torch.nn import Flatten

from backpack.core.derivatives.basederivatives import BaseDerivatives


class FlattenDerivatives(BaseDerivatives):
    def get_module(self):
        return Flatten

    def hessian_is_zero(self):
        return True

    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        return mat

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        return self.reshape_like_input(mat, module)

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        return self.reshape_like_output(mat, module)

    def is_no_op(self, module):
        """Does flatten add an operation to the computational graph.

        If the input is already flattened, no operation will be added for
        the `Flatten` layer. This can lead to an intuitive order of backward
        hook execution, see the discussion at https://discuss.pytorch.org/t/
        backward-hooks-changing-order-of-execution-in-nn-sequential/12447/4 .
        """
        return tuple(module.input0_shape) == tuple(module.output_shape)



from torch.nn import Conv2d, Linear

from backpack.extensions.backprop_extension import BackpropExtension

from . import conv2d, linear


class SumGradSquared(BackpropExtension):
    """The sum of individual-gradients-squared, or second moment of the gradient.

    Stores the output in ``sum_grad_squared``. Same dimension as the gradient.

    Note: beware of scaling issue
        The second moment depends on the scaling of the overall function.
        Let ``fᵢ`` be the loss of the ``i`` th sample, with gradient ``gᵢ``.
        ``SumGradSquared`` will return the sum of the squared

        - ``[g₁, …, gₙ]`` if the loss is a sum, ``∑ᵢ₌₁ⁿ fᵢ``,
        - ``[¹/ₙ g₁, …, ¹/ₙ gₙ]`` if the loss is a mean, ``¹/ₙ ∑ᵢ₌₁ⁿ fᵢ``.
    """

    def __init__(self):
        super().__init__(
            savefield="sum_grad_squared",
            fail_mode="WARNING",
            module_exts={Linear: linear.SGSLinear(), Conv2d: conv2d.SGSConv2d(),},
        )



from warnings import warn

from torch import einsum
from torch.nn import BatchNorm1d

from backpack.core.derivatives.basederivatives import BaseParameterDerivatives
from backpack.core.derivatives.shape_check import (
    R_mat_prod_accept_vectors,
    R_mat_prod_check_shapes,
)


class BatchNorm1dDerivatives(BaseParameterDerivatives):
    def get_module(self):
        return BatchNorm1d

    def hessian_is_zero(self):
        return False

    def hessian_is_diagonal(self):
        return False

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        return self._jac_t_mat_prod(module, g_inp, g_out, mat)

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        """
        Note:
        -----
        The Jacobian is *not independent* among the batch dimension, i.e.
        D z_i = D z_i(x_1, ..., x_B).

        This structure breaks the computation of the GGN diagonal,
        for curvature-matrix products it should still work.

        References:
        -----------
        https://kevinzakka.github.io/2016/09/14/batch_normalization/
        https://chrisyeh96.github.io/2017/08/28/deriving-batchnorm-backprop.html
        """
        assert module.affine is True

        N = self.get_batch(module)
        x_hat, var = self.get_normalized_input_and_var(module)
        ivar = 1.0 / (var + module.eps).sqrt()

        dx_hat = einsum("vni,i->vni", (mat, module.weight))

        jac_t_mat = N * dx_hat
        jac_t_mat -= dx_hat.sum(1).unsqueeze(1).expand_as(jac_t_mat)
        jac_t_mat -= einsum("ni,vsi,si->vni", (x_hat, dx_hat, x_hat))
        jac_t_mat = einsum("vni,i->vni", (jac_t_mat, ivar / N))

        return jac_t_mat

    def get_normalized_input_and_var(self, module):
        input = module.input0
        mean = input.mean(dim=0)
        var = input.var(dim=0, unbiased=False)
        return (input - mean) / (var + module.eps).sqrt(), var

    @R_mat_prod_accept_vectors
    @R_mat_prod_check_shapes
    def make_residual_mat_prod(self, module, g_inp, g_out):
        # TODO: Implement R_mat_prod for BatchNorm
        def R_mat_prod(mat):
            """Multiply with the residual: mat → [∑_{k} Hz_k(x) 𝛿z_k] mat.

            Second term of the module input Hessian backpropagation equation.
            """
            raise NotImplementedError

        # TODO: Enable tests in test/automated_bn_test.py
        raise NotImplementedError
        return R_mat_prod

    def _weight_jac_mat_prod(self, module, g_inp, g_out, mat):
        x_hat, _ = self.get_normalized_input_and_var(module)
        return einsum("ni,vi->vni", (x_hat, mat))

    def _weight_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch):
        if not sum_batch:
            warn(
                "BatchNorm batch summation disabled."
                "This may not compute meaningful quantities"
            )
        x_hat, _ = self.get_normalized_input_and_var(module)
        equation = "vni,ni->v{}i".format("" if sum_batch is True else "n")
        operands = [mat, x_hat]
        return einsum(equation, operands)

    def _bias_jac_mat_prod(self, module, g_inp, g_out, mat):
        N = self.get_batch(module)
        return mat.unsqueeze(1).repeat(1, N, 1)

    def _bias_jac_t_mat_prod(self, module, g_inp, g_out, mat, sum_batch=True):
        if not sum_batch:
            warn(
                "BatchNorm batch summation disabled."
                "This may not compute meaningful quantities"
            )
            return mat
        else:
            N_axis = 1
            return mat.sum(N_axis)



from torch import einsum

from backpack.extensions.firstorder.base import FirstOrderModuleExtension
from backpack.utils import conv as convUtils


class BatchL2Conv2d(FirstOrderModuleExtension):
    def __init__(self):
        super().__init__(params=["bias", "weight"])

    def bias(self, ext, module, g_inp, g_out, backproped):
        C_axis = 1
        return (einsum("nchw->nc", g_out[0]) ** 2).sum(C_axis)

    def weight(self, ext, module, g_inp, g_out, backproped):
        X, dE_dY = convUtils.get_weight_gradient_factors(
            module.input0, g_out[0], module
        )
        return einsum("nml,nkl,nmi,nki->n", (dE_dY, X, dE_dY, X))



import warnings


class ModuleExtension:
    """
    Base class for a Module Extension for BackPACK.

    Descendants of this class need to
    - define what parameters of the Module need to be treated (weight, bias)
      and provide functions to compute the quantities
    - extend the `backpropagate` function if information other than the gradient
      needs to be propagated through the graph.
    """

    def __init__(self, params=None):
        """
        Parameters
        ----------
        params: [str]
            List of module parameters that need special treatment.
            for each param `p` in the list, instances of the extended module `m`
            need to have a field `m.p` and the class extending `ModuleExtension`
            need to provide a method with the same signature as the `backprop`
            method.
            The result of this method will be saved in the savefield of `m.p`.
        """
        if params is None:
            params = []

        self.__params = params

        for param in self.__params:
            extFunc = getattr(self, param, None)
            if extFunc is None:
                raise NotImplementedError

    def backpropagate(self, ext, module, g_inp, g_out, bpQuantities):
        """
        Main method to extend to backpropagate additional information through
        the graph.

        Parameters
        ----------
        ext: BackpropExtension
            Instance of the extension currently running
        module: torch.nn.Module
            Instance of the extended module
        g_inp: [Tensor]
            Gradient of the loss w.r.t. the inputs
        g_out: Tensor
            Gradient of the loss w.r.t. the output
        bpQuantities:
            Quantities backpropagated w.r.t. the output

        Returns
        -------
        bpQuantities:
            Quantities backpropagated w.r.t. the input
        """
        warnings.warn("Backpropagate has not been overwritten")

    def apply(self, ext, module, g_inp, g_out):
        inp = module.input0
        out = module.output

        bpQuantities = self.__backproped_quantities(ext, out)

        for param in self.__params:
            if self.__param_exists_and_requires_grad(module, param):
                extFunc = getattr(self, param)
                extValue = extFunc(ext, module, g_inp, g_out, bpQuantities)
                self.__save(extValue, ext, module, param)

        bpQuantities = self.backpropagate(ext, module, g_inp, g_out, bpQuantities)

        self.__backprop_quantities(ext, inp, out, bpQuantities)

    @staticmethod
    def __backproped_quantities(ext, out):
        return getattr(out, ext.savefield, None)

    @staticmethod
    def __backprop_quantities(ext, inp, out, bpQuantities):

        setattr(inp, ext.savefield, bpQuantities)

        is_a_leaf = out.grad_fn is None
        retain_grad_is_on = getattr(out, "retains_grad", False)
        inp_is_out = id(inp) == id(out)
        should_retain_grad = is_a_leaf or retain_grad_is_on or inp_is_out

        if not should_retain_grad:
            if hasattr(out, ext.savefield):
                delattr(out, ext.savefield)

    @staticmethod
    def __param_exists_and_requires_grad(module, param):
        param_exists = getattr(module, param) is not None
        return param_exists and getattr(module, param).requires_grad

    @staticmethod
    def __save(value, extension, module, param):
        setattr(getattr(module, param), extension.savefield, value)



"""Utility functions for examples."""
import torch
import torchvision


def load_mnist_dataset():
    """Download and normalize MNIST training data."""
    mnist_dataset = torchvision.datasets.MNIST(
        root="./data",
        train=True,
        transform=torchvision.transforms.Compose(
            [
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize((0.1307,), (0.3081,)),
            ]
        ),
        download=True,
    )
    return mnist_dataset


def get_mnist_dataloder(batch_size=64, shuffle=True):
    """Returns a dataloader for MNIST"""
    return torch.utils.data.dataloader.DataLoader(
        load_mnist_dataset(), batch_size=batch_size, shuffle=shuffle,
    )


def load_one_batch_mnist(batch_size=64, shuffle=True):
    """Return a single batch (inputs, labels) of MNIST data."""
    dataloader = get_mnist_dataloder(batch_size, shuffle)
    X, y = next(iter(dataloader))
    return X, y



from os import path

from setuptools import find_packages, setup

# META
##############################################################################
AUTHORS = "F. Dangel, F. Kunstner"
NAME = "backpack-for-pytorch"
PACKAGES = find_packages()

DESCRIPTION = "BackPACK: Packing more into backprop"
LONG_DESCR = """
    BackPACK is built on top of PyTorch.
    It efficiently computes quantities other than the gradient.

    Website: https://backpack.pt
    Code: https://github.com/f-dangel/backpack
    Documentation: https://readthedocs.org/projects/backpack/
    Bug reports & feature requests: https://github.com/f-dangel/backpack/issues
    """

VERSION = "1.1.1"
URL = "https://github.com/f-dangel/backpack"
LICENSE = "MIT"

# DEPENDENCIES
##############################################################################
REQUIREMENTS_FILE = "requirements.txt"
REQUIREMENTS_PATH = path.join(path.abspath(__file__), REQUIREMENTS_FILE)

with open(REQUIREMENTS_FILE) as f:
    requirements = f.read().splitlines()

setup(
    author=AUTHORS,
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCR,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    url=URL,
    license=LICENSE,
    packages=PACKAGES,
    zip_safe=False,
    python_requires=">=3.5",
)



from backpack.core.derivatives.zeropad2d import ZeroPad2dDerivatives

from .cmpbase import CMPBase


class CMPZeroPad2d(CMPBase):
    def __init__(self):
        super().__init__(derivatives=ZeroPad2dDerivatives())



from torch import gt
from torch.nn import ReLU

from backpack.core.derivatives.elementwise import ElementwiseDerivatives


class ReLUDerivatives(ElementwiseDerivatives):
    def get_module(self):
        return ReLU

    def hessian_is_zero(self):
        return True

    def df(self, module, g_inp, g_out):
        return gt(module.input0, 0).float()



from backpack.core.derivatives.relu import ReLUDerivatives
from backpack.core.derivatives.sigmoid import SigmoidDerivatives
from backpack.core.derivatives.tanh import TanhDerivatives
from backpack.extensions.secondorder.diag_hessian.diag_h_base import DiagHBaseModule


class DiagHReLU(DiagHBaseModule):
    def __init__(self):
        super().__init__(derivatives=ReLUDerivatives())


class DiagHSigmoid(DiagHBaseModule):
    def __init__(self):
        super().__init__(derivatives=SigmoidDerivatives())


class DiagHTanh(DiagHBaseModule):
    def __init__(self):
        super().__init__(derivatives=TanhDerivatives())



from backpack.core.derivatives.batchnorm1d import BatchNorm1dDerivatives

from .base import GradBaseModule


class GradBatchNorm1d(GradBaseModule):
    def __init__(self):
        super().__init__(
            derivatives=BatchNorm1dDerivatives(), params=["bias", "weight"]
        )



from backpack.core.derivatives.linear import LinearDerivatives
from backpack.extensions.curvmatprod.cmpbase import CMPBase


class CMPLinear(CMPBase):
    def __init__(self):
        super().__init__(derivatives=LinearDerivatives(), params=["weight", "bias"])

    def weight(self, ext, module, g_inp, g_out, backproped):
        CMP_out = backproped

        def weight_cmp(mat):
            Jmat = self.derivatives.weight_jac_mat_prod(module, g_inp, g_out, mat)
            CJmat = CMP_out(Jmat)
            JTCJmat = self.derivatives.weight_jac_t_mat_prod(
                module, g_inp, g_out, CJmat
            )
            return JTCJmat

        return weight_cmp

    def bias(self, ext, module, g_inp, g_out, backproped):
        CMP_out = backproped

        def bias_cmp(mat):
            Jmat = self.derivatives.bias_jac_mat_prod(module, g_inp, g_out, mat)
            CJmat = CMP_out(Jmat)
            JTCJmat = self.derivatives.bias_jac_t_mat_prod(module, g_inp, g_out, CJmat)
            return JTCJmat

        return bias_cmp



from backpack.core.derivatives.relu import ReLUDerivatives
from backpack.core.derivatives.sigmoid import SigmoidDerivatives
from backpack.core.derivatives.tanh import TanhDerivatives

from .cmpbase import CMPBase


class CMPReLU(CMPBase):
    def __init__(self):
        super().__init__(derivatives=ReLUDerivatives())


class CMPSigmoid(CMPBase):
    def __init__(self):
        super().__init__(derivatives=SigmoidDerivatives())


class CMPTanh(CMPBase):
    def __init__(self):
        super().__init__(derivatives=TanhDerivatives())



"""Partial derivatives for cross-entropy loss."""
from math import sqrt

from torch import diag, diag_embed, einsum, multinomial, ones_like, softmax
from torch import sqrt as torchsqrt
from torch.nn import CrossEntropyLoss
from torch.nn.functional import one_hot

from backpack.core.derivatives.basederivatives import BaseLossDerivatives


class CrossEntropyLossDerivatives(BaseLossDerivatives):
    """Partial derivatives for cross-entropy loss.

    The `torch.nn.CrossEntropyLoss` operation is a composition of softmax
    and negative log-likelihood.
    """

    def get_module(self):
        """Return the `torch.nn` module for cross-entropy loss."""
        return CrossEntropyLoss

    def _sqrt_hessian(self, module, g_inp, g_out):
        self._check_2nd_order_parameters(module)

        probs = self._get_probs(module)
        tau = torchsqrt(probs)
        V_dim, C_dim = 0, 2
        Id = diag_embed(ones_like(probs), dim1=V_dim, dim2=C_dim)
        Id_tautau = Id - einsum("nv,nc->vnc", tau, tau)
        sqrt_H = einsum("nc,vnc->vnc", tau, Id_tautau)

        if module.reduction == "mean":
            N = module.input0.shape[0]
            sqrt_H /= sqrt(N)

        return sqrt_H

    def _sqrt_hessian_sampled(self, module, g_inp, g_out, mc_samples=1):
        self._check_2nd_order_parameters(module)

        M = mc_samples
        C = module.input0.shape[1]

        probs = self._get_probs(module)
        V_dim = 0
        probs_unsqueezed = probs.unsqueeze(V_dim).repeat(M, 1, 1)

        multi = multinomial(probs, M, replacement=True)
        classes = one_hot(multi, num_classes=C)
        classes = einsum("nvc->vnc", classes).float()

        sqrt_mc_h = (probs_unsqueezed - classes) / sqrt(M)

        if module.reduction == "mean":
            N = module.input0.shape[0]
            sqrt_mc_h /= sqrt(N)

        return sqrt_mc_h

    def _sum_hessian(self, module, g_inp, g_out):
        self._check_2nd_order_parameters(module)

        probs = self._get_probs(module)
        sum_H = diag(probs.sum(0)) - einsum("bi,bj->ij", (probs, probs))

        if module.reduction == "mean":
            N = module.input0.shape[0]
            sum_H /= N

        return sum_H

    def _make_hessian_mat_prod(self, module, g_inp, g_out):
        """Multiplication of the input Hessian with a matrix."""
        self._check_2nd_order_parameters(module)

        probs = self._get_probs(module)

        def hessian_mat_prod(mat):
            Hmat = einsum("bi,cbi->cbi", (probs, mat)) - einsum(
                "bi,bj,cbj->cbi", (probs, probs, mat)
            )

            if module.reduction == "mean":
                N = module.input0.shape[0]
                Hmat /= N

            return Hmat

        return hessian_mat_prod

    def hessian_is_psd(self):
        """Return whether cross-entropy loss Hessian is positive semi-definite."""
        return True

    def _get_probs(self, module):
        return softmax(module.input0, dim=1)

    def _check_2nd_order_parameters(self, module):
        """Verify that the parameters are supported by 2nd-order quantities.

        Attributes:
            module (torch.nn.CrossEntropyLoss): Extended CrossEntropyLoss module

        Raises:
            NotImplementedError: If module's setting is not implemented.
        """
        implemented_ignore_index = -100
        implemented_weight = None

        if module.ignore_index != implemented_ignore_index:
            raise NotImplementedError(
                "Only default ignore_index ({}) is implemented, got {}".format(
                    implemented_ignore_index, module.ignore_index
                )
            )

        if module.weight != implemented_weight:
            raise NotImplementedError(
                "Only default weight ({}) is implemented, got {}".format(
                    implemented_weight, module.weight
                )
            )



from torch import zeros
from torch.nn import MaxPool2d
from torch.nn.functional import max_pool2d

from backpack.core.derivatives.basederivatives import BaseDerivatives
from backpack.utils.ein import eingroup


class MaxPool2DDerivatives(BaseDerivatives):
    def get_module(self):
        return MaxPool2d

    # TODO: Do not recompute but get from forward pass of module
    def get_pooling_idx(self, module):
        _, pool_idx = max_pool2d(
            module.input0,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            return_indices=True,
            ceil_mode=module.ceil_mode,
        )
        return pool_idx

    def ea_jac_t_mat_jac_prod(self, module, g_inp, g_out, mat):
        """

        Note: It is highly questionable whether this makes sense both
              in terms of the approximation and memory costs.

        Note:
            Need to loop over the samples, as dealing with all at once
            requires memory for `N * C² * H_in * W_in * H_out * W_out`
            elements
        """
        device = mat.device
        N, C, H_in, W_in = module.input0.size()
        _, _, H_out, W_out = module.output.size()

        in_pixels = H_in * W_in
        in_features = C * in_pixels

        pool_idx = self.get_pooling_idx(module).view(N, C, H_out * W_out)

        def sample_ea_jac_t_mat_jac_prod(n, mat):
            jac_t_mat = sample_jac_t_mat_prod(n, mat)
            mat_t_jac = jac_t_mat.t()
            jac_t_mat_t_jac = sample_jac_t_mat_prod(n, mat_t_jac)
            return jac_t_mat_t_jac.t()

        def sample_jac_t_mat_prod(n, mat):
            num_cols = mat.size(1)
            idx = pool_idx[n, :, :].unsqueeze(-1).expand(-1, -1, num_cols)

            jac_t_mat = zeros(C, H_in * W_in, num_cols, device=device)
            mat = mat.reshape(C, H_out * W_out, num_cols)

            jac_t_mat.scatter_add_(1, idx, mat)

            return jac_t_mat.reshape(in_features, num_cols)

        result = zeros(in_features, in_features, device=device)

        for n in range(N):
            result += sample_ea_jac_t_mat_jac_prod(n, mat)

        return result / N

    def hessian_is_zero(self):
        return True

    def _jac_mat_prod(self, module, g_inp, g_out, mat):
        mat_as_pool = eingroup("v,n,c,h,w->v,n,c,hw", mat)
        jmp_as_pool = self.__apply_jacobian_of(module, mat_as_pool)
        return self.reshape_like_output(jmp_as_pool, module)

    def __apply_jacobian_of(self, module, mat):
        V, HW_axis = mat.shape[0], 3
        pool_idx = self.__pool_idx_for_jac(module, V)
        return mat.gather(HW_axis, pool_idx)

    def __pool_idx_for_jac(self, module, V):
        """Manipulated pooling indices ready-to-use in jac(t)."""

        pool_idx = self.get_pooling_idx(module)
        V_axis = 0
        return (
            eingroup("n,c,h,w->n,c,hw", pool_idx)
            .unsqueeze(V_axis)
            .expand(V, -1, -1, -1)
        )

    def _jac_t_mat_prod(self, module, g_inp, g_out, mat):
        mat_as_pool = eingroup("v,n,c,h,w->v,n,c,hw", mat)
        jmp_as_pool = self.__apply_jacobian_t_of(module, mat_as_pool)
        return self.reshape_like_input(jmp_as_pool, module)

    def __apply_jacobian_t_of(self, module, mat):
        V = mat.shape[0]
        result = self.__zero_for_jac_t(module, V, mat.device)
        pool_idx = self.__pool_idx_for_jac(module, V)

        HW_axis = 3
        result.scatter_add_(HW_axis, pool_idx, mat)
        return result

    def __zero_for_jac_t(self, module, V, device):
        N, C_out, _, _ = module.output_shape
        _, _, H_in, W_in = module.input0.size()

        shape = (V, N, C_out, H_in * W_in)
        return zeros(shape, device=device)



""" Einsum utility functions. """

import torch


def eingroup(equation, operand, dim=None):
    """Use einsum-like notation for (un-)grouping dimensions.

    Dimensions that cannot be inferred can be handed in via a mapping `dim`.

    Arguments:
        equation (str): Equation specifying the (un-)grouping of axes.
        operand (torch.Tensor): The tensor that `equation` will be applied to.
        dim (dict, optional): A mapping from letters in `equation` to
            dimensions. Only required if `eingroup` cannot infer the dimension.
            For instance, consider you want to interpret a vector with 10
            elements as a 5x2 matrix. The equation `"i,j->ij"` is not
            sufficient, you need to specify `dim = {"i": 5, "j": 2}`.

    Note:
        Many operations in `backpack` require that certain axes of a tensor
        be treated identically, and will therefore be grouped into a single
        dimension. One way to do that is using `view`s or `reshape`s.
        `eingroup` helps facilitate this process. It can be used in roughly
        the same way as `einsum`, but acts only on a single tensor at
        a time (although this could be fixed with an improved syntax and
        equation analysis).

        Idea:
        * `"a,b,c->ab,c"`: group dimension `a` and `b` into a single one.
        * `"a,b,c->ba,c"` to transpose, then group `b` and `a` dimension.

    Examples:
        Different reshapes of a [2 x 2 x 2] tensor:
        >>> t = torch.Tensor([0, 1, 2, 3, 4, 5, 6, 7]).reshape(2, 2, 2)
        >>> t_flat = t.reshape(-1)
        >>> # group all dimensions
        >>> eingroup("i,j,k->ijk", t)
        torch.Tensor([0, 1, 2, 3, 4, 5, 6, 7])
        >>> # interpret as 2 x 4 matrix
        >>> eingroup("i,j,k->i,jk", t)
        torch.Tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        >>> # grouping (specifying grouping dimensions)
        >>> eingroup("ijk->i,j,k", dim={"i": 2, "j": 2, "k": 2})
        torch.Tensor([[[0, 1], [2, 3]], [[4, 5], [6, 7]]])
        >>> # grouping with additional contraction
        >>> eingroup("ijk->j,k", dim={"j": 2, "k": 2})
        torch.Tensor([[[4, 5], [8, 10]])

    Returns:
        torch.Tensor: Result of the (un-)grouping operation.

    Raises:
        KeyError: If information about a dimension in `dim` is missing
            or can be removed. # noqa: DAR402
        RuntimeError: If the groups inferred from `equation` do not match
            the number of axes of `operand` # noqa: DAR402

    """
    dim = {} if dim is None else dim
    in_shape, out_shape, einsum_eq = _eingroup_preprocess(equation, operand, dim=dim)

    operand_in = operand.reshape(in_shape)
    result = torch.einsum(einsum_eq, operand_in)
    return result.reshape(out_shape)


def _eingroup_preprocess(equation, operand, dim):
    """Process `eingroup` equation.

    Return the `reshape`s and `einsum` equations that have to
    be performed.
    """
    split, sep = "->", ","

    def groups(string):
        return string.split(sep)

    lhs, rhs = equation.split(split)
    in_groups, out_groups = groups(lhs), groups(rhs)

    dim = __eingroup_infer(in_groups, operand, dim)
    in_shape_flat, out_shape = __eingroup_shapes(in_groups, out_groups, dim)

    return in_shape_flat, out_shape, equation.replace(sep, "")


def __eingroup_shapes(in_groups, out_groups, dim):
    """Return shape the input needs to be reshaped, and the output shape."""

    def shape(groups, dim):
        return [group_dim(group, dim) for group in groups]

    def product(nums):
        assert len(nums) > 0

        result = 1
        for num in nums:
            result *= num
        return result

    def group_dim(group, dim):
        try:
            return product([dim[g] for g in group])
        except KeyError as e:
            raise KeyError("Unknown dimension for an axis {}".format(e))

    out_shape = shape(out_groups, dim)

    in_groups_flat = []
    for group in in_groups:
        for letter in group:
            in_groups_flat.append(letter)
    in_shape_flat = shape(in_groups_flat, dim)

    return in_shape_flat, out_shape


def __eingroup_infer(in_groups, operand, dim):
    """Infer the size of each axis."""
    if not len(in_groups) == len(operand.shape):
        raise RuntimeError(
            "Got {} input groups {}, but tensor has {} axes.".format(
                len(in_groups), in_groups, len(operand.shape)
            )
        )

    for group, size in zip(in_groups, operand.shape):
        if len(group) == 1:
            axis = group[0]
            if axis in dim.keys():
                raise KeyError(
                    "Can infer dimension of axis {}.".format(axis),
                    "Remove from dim = {}.".format(dim),
                )
            dim[axis] = size

    return dim



"""
BackPACK Extensions
"""

from .curvmatprod import CMP
from .firstorder import BatchGrad, BatchL2Grad, SumGradSquared, Variance
from .secondorder import (
    KFAC,
    KFLR,
    KFRA,
    HBP,
    DiagGGNExact,
    DiagGGNMC,
    DiagGGN,
    DiagHessian,
)

__all__ = [
    "CMP",
    "BatchL2Grad",
    "BatchGrad",
    "SumGradSquared",
    "Variance",
    "KFAC",
    "KFLR",
    "KFRA",
    "HBP",
    "DiagGGNExact",
    "DiagGGNMC",
    "DiagGGN",
    "DiagHessian",
]
